import asyncio
import json
from abc import abstractmethod
from decimal import Decimal
from typing import Callable, List, Optional, Tuple
from unittest.mock import AsyncMock, patch

from aioresponses import aioresponses

from hummingbot.connector.derivative.position import Position
from hummingbot.connector.test_support.exchange_connector_test import AbstractExchangeConnectorTests
from hummingbot.core.data_type.common import OrderType, PositionAction, PositionMode, PositionSide, TradeType
from hummingbot.core.data_type.funding_info import FundingInfo
from hummingbot.core.data_type.in_flight_order import InFlightOrder
from hummingbot.core.event.event_logger import EventLogger
from hummingbot.core.event.events import (
    BuyOrderCompletedEvent,
    BuyOrderCreatedEvent,
    FundingPaymentCompletedEvent,
    MarketEvent,
    OrderFilledEvent,
    SellOrderCompletedEvent,
    SellOrderCreatedEvent,
)


class AbstractPerpetualDerivativeTests:
    """
    We need to create the abstract TestCase class inside another class not inheriting from TestCase to prevent test
    frameworks from discovering and tyring to run the abstract class
    """

    class PerpetualDerivativeTests(AbstractExchangeConnectorTests.ExchangeConnectorTests):
        @property
        @abstractmethod
        def expected_supported_position_modes(self) -> List[PositionMode]:
            raise NotImplementedError

        @property
        @abstractmethod
        def funding_info_url(self):
            raise NotImplementedError

        @property
        @abstractmethod
        def funding_payment_url(self):
            raise NotImplementedError

        @property
        @abstractmethod
        def funding_info_mock_response(self):
            raise NotImplementedError

        @property
        @abstractmethod
        def empty_funding_payment_mock_response(self):
            raise NotImplementedError

        @property
        @abstractmethod
        def funding_payment_mock_response(self):
            raise NotImplementedError

        @property
        def target_funding_info_index_price(self):
            return 1

        @property
        def target_funding_info_mark_price(self):
            return 2

        @property
        def target_funding_info_next_funding_utc_timestamp(self):
            return 1657099053

        @property
        def target_funding_info_rate(self):
            return 3

        @property
        def target_funding_info_index_price_ws_updated(self):
            return 10

        @property
        def target_funding_info_mark_price_ws_updated(self):
            return 20

        @property
        def target_funding_info_next_funding_utc_timestamp_ws_updated(self):
            return 1657100053

        @property
        def target_funding_info_rate_ws_updated(self):
            return 30

        @property
        def target_funding_payment_timestamp(self):
            return 1657110053

        @property
        def target_funding_payment_funding_rate(self):
            return 100

        @property
        def target_funding_payment_payment_amount(self):
            return 200

        @abstractmethod
        def position_event_for_full_fill_websocket_update(self, order: InFlightOrder, unrealized_pnl: float):
            raise NotImplementedError

        @abstractmethod
        def configure_successful_set_position_mode(
            self,
            position_mode: PositionMode,
            mock_api: aioresponses,
            callback: Optional[Callable] = lambda *args, **kwargs: None,
        ):
            raise NotImplementedError

        @abstractmethod
        def configure_failed_set_position_mode(
            self,
            position_mode: PositionMode,
            mock_api: aioresponses,
            callback: Optional[Callable] = lambda *args, **kwargs: None,
        ) -> Tuple[str, str]:
            """
            :return: A tuple of the URL and an error message if the exchange returns one on failure.
            """
            raise NotImplementedError

        @abstractmethod
        def configure_failed_set_leverage(
            self,
            leverage: int,
            mock_api: aioresponses,
            callback: Optional[Callable] = lambda *args, **kwargs: None,
        ) -> Tuple[str, str]:
            """
            :return: A tuple of the URL and an error message if the exchange returns one on failure.
            """
            raise NotImplementedError

        @abstractmethod
        def configure_successful_set_leverage(
            self,
            leverage: int,
            mock_api: aioresponses,
            callback: Optional[Callable] = lambda *args, **kwargs: None,
        ):
            raise NotImplementedError

        @abstractmethod
        def funding_info_event_for_websocket_update(self):
            raise NotImplementedError

        def place_buy_order(
            self,
            amount: Decimal = Decimal("100"),
            price: Decimal = Decimal("10_000"),
            order_type: OrderType = OrderType.LIMIT,
            position_action: PositionAction = PositionAction.OPEN,
        ):
            order_id = self.exchange.buy(
                trading_pair=self.trading_pair,
                amount=amount,
                order_type=order_type,
                price=price,
                position_action=position_action,
            )
            return order_id

        def place_sell_order(
            self,
            amount: Decimal = Decimal("100"),
            price: Decimal = Decimal("10_000"),
            order_type: OrderType = OrderType.LIMIT,
            position_action: PositionAction = PositionAction.OPEN,
        ):
            order_id = self.exchange.sell(
                trading_pair=self.trading_pair,
                amount=amount,
                order_type=order_type,
                price=price,
                position_action=position_action,
            )
            return order_id

        def _initialize_event_loggers(self):
            super()._initialize_event_loggers()
            self.funding_payment_logger = EventLogger()
            self.exchange.add_listener(MarketEvent.FundingPaymentCompleted, self.funding_payment_logger)

        def test_initial_status_dict(self):
            self.exchange._set_trading_pair_symbol_map(None)
            self.exchange._perpetual_trading._funding_info = {}

            status_dict = self.exchange.status_dict

            expected_initial_dict = self._expected_initial_status_dict()
            expected_initial_dict["funding_info"] = False

            self.assertEqual(expected_initial_dict, status_dict)
            self.assertFalse(self.exchange.ready)

        @aioresponses()
        def test_create_buy_limit_order_successfully(self, mock_api):
            """Open long position"""
            self._simulate_trading_rules_initialized()
            request_sent_event = asyncio.Event()
            self.exchange._set_current_timestamp(1640780000)

            url = self.order_creation_url

            creation_response = self.order_creation_request_successful_mock_response

            mock_api.post(url,
                          body=json.dumps(creation_response),
                          callback=lambda *args, **kwargs: request_sent_event.set())

            leverage = 2
            self.exchange._perpetual_trading.set_leverage(self.trading_pair, leverage)
            order_id = self.place_buy_order()
            self.async_run_with_timeout(request_sent_event.wait())

            order_request = self._all_executed_requests(mock_api, url)[0]
            self.validate_auth_credentials_present(order_request)
            self.assertIn(order_id, self.exchange.in_flight_orders)
            self.validate_order_creation_request(
                order=self.exchange.in_flight_orders[order_id],
                request_call=order_request)

            create_event: BuyOrderCreatedEvent = self.buy_order_created_logger.event_log[0]
            self.assertEqual(self.exchange.current_timestamp,
                             create_event.timestamp)
            self.assertEqual(self.trading_pair, create_event.trading_pair)
            self.assertEqual(OrderType.LIMIT, create_event.type)
            self.assertEqual(Decimal("100"), create_event.amount)
            self.assertEqual(Decimal("10000"), create_event.price)
            self.assertEqual(order_id, create_event.order_id)
            self.assertEqual(str(self.expected_exchange_order_id),
                             create_event.exchange_order_id)
            self.assertEqual(leverage, create_event.leverage)
            self.assertEqual(PositionAction.OPEN.value, create_event.position)

            self.assertTrue(
                self.is_logged(
                    "INFO",
                    f"Created {OrderType.LIMIT.name} {TradeType.BUY.name} order {order_id} for "
                    f"{Decimal('100.000000')} to {PositionAction.OPEN.name} a {self.trading_pair} position "
                    f"at {Decimal('10000.0000')}."
                )
            )

        @aioresponses()
        def test_create_sell_limit_order_successfully(self, mock_api):
            """Open short position"""
            self._simulate_trading_rules_initialized()
            request_sent_event = asyncio.Event()
            self.exchange._set_current_timestamp(1640780000)

            url = self.order_creation_url
            creation_response = self.order_creation_request_successful_mock_response

            mock_api.post(url,
                          body=json.dumps(creation_response),
                          callback=lambda *args, **kwargs: request_sent_event.set())
            leverage = 3
            self.exchange._perpetual_trading.set_leverage(self.trading_pair, leverage)
            order_id = self.place_sell_order()
            self.async_run_with_timeout(request_sent_event.wait())

            order_request = self._all_executed_requests(mock_api, url)[0]
            self.validate_auth_credentials_present(order_request)
            self.assertIn(order_id, self.exchange.in_flight_orders)
            self.validate_order_creation_request(
                order=self.exchange.in_flight_orders[order_id],
                request_call=order_request)

            create_event: SellOrderCreatedEvent = self.sell_order_created_logger.event_log[0]
            self.assertEqual(self.exchange.current_timestamp, create_event.timestamp)
            self.assertEqual(self.trading_pair, create_event.trading_pair)
            self.assertEqual(OrderType.LIMIT, create_event.type)
            self.assertEqual(Decimal("100"), create_event.amount)
            self.assertEqual(Decimal("10000"), create_event.price)
            self.assertEqual(order_id, create_event.order_id)
            self.assertEqual(str(self.expected_exchange_order_id), create_event.exchange_order_id)
            self.assertEqual(leverage, create_event.leverage)
            self.assertEqual(PositionAction.OPEN.value, create_event.position)

            self.assertTrue(
                self.is_logged(
                    "INFO",
                    f"Created {OrderType.LIMIT.name} {TradeType.SELL.name} order {order_id} for "
                    f"{Decimal('100.000000')} to {PositionAction.OPEN.name} a {self.trading_pair} position "
                    f"at {Decimal('10000.0000')}."
                )
            )

        @aioresponses()
        def test_create_order_to_close_short_position(self, mock_api):
            self._simulate_trading_rules_initialized()
            request_sent_event = asyncio.Event()
            self.exchange._set_current_timestamp(1640780000)

            url = self.order_creation_url

            creation_response = self.order_creation_request_successful_mock_response

            mock_api.post(url,
                          body=json.dumps(creation_response),
                          callback=lambda *args, **kwargs: request_sent_event.set())
            leverage = 4
            self.exchange._perpetual_trading.set_leverage(self.trading_pair, leverage)
            order_id = self.place_buy_order(position_action=PositionAction.CLOSE)
            self.async_run_with_timeout(request_sent_event.wait())

            order_request = self._all_executed_requests(mock_api, url)[0]
            self.validate_auth_credentials_present(order_request)
            self.assertIn(order_id, self.exchange.in_flight_orders)
            self.validate_order_creation_request(
                order=self.exchange.in_flight_orders[order_id],
                request_call=order_request)

            create_event: BuyOrderCreatedEvent = self.buy_order_created_logger.event_log[0]
            self.assertEqual(self.exchange.current_timestamp,
                             create_event.timestamp)
            self.assertEqual(self.trading_pair, create_event.trading_pair)
            self.assertEqual(OrderType.LIMIT, create_event.type)
            self.assertEqual(Decimal("100"), create_event.amount)
            self.assertEqual(Decimal("10000"), create_event.price)
            self.assertEqual(order_id, create_event.order_id)
            self.assertEqual(str(self.expected_exchange_order_id),
                             create_event.exchange_order_id)
            self.assertEqual(leverage, create_event.leverage)
            self.assertEqual(PositionAction.CLOSE.value, create_event.position)

            self.assertTrue(
                self.is_logged(
                    "INFO",
                    f"Created {OrderType.LIMIT.name} {TradeType.BUY.name} order {order_id} for "
                    f"{Decimal('100.000000')} to {PositionAction.CLOSE.name} a {self.trading_pair} position "
                    f"at {Decimal('10000.0000')}."
                )
            )

        @aioresponses()
        def test_create_order_to_close_long_position(self, mock_api):
            self._simulate_trading_rules_initialized()
            request_sent_event = asyncio.Event()
            self.exchange._set_current_timestamp(1640780000)

            url = self.order_creation_url
            creation_response = self.order_creation_request_successful_mock_response

            mock_api.post(url,
                          body=json.dumps(creation_response),
                          callback=lambda *args, **kwargs: request_sent_event.set())
            leverage = 5
            self.exchange._perpetual_trading.set_leverage(self.trading_pair, leverage)
            order_id = self.place_sell_order(position_action=PositionAction.CLOSE)
            self.async_run_with_timeout(request_sent_event.wait())

            order_request = self._all_executed_requests(mock_api, url)[0]
            self.validate_auth_credentials_present(order_request)
            self.assertIn(order_id, self.exchange.in_flight_orders)
            self.validate_order_creation_request(
                order=self.exchange.in_flight_orders[order_id],
                request_call=order_request)

            create_event: SellOrderCreatedEvent = self.sell_order_created_logger.event_log[0]
            self.assertEqual(self.exchange.current_timestamp, create_event.timestamp)
            self.assertEqual(self.trading_pair, create_event.trading_pair)
            self.assertEqual(OrderType.LIMIT, create_event.type)
            self.assertEqual(Decimal("100"), create_event.amount)
            self.assertEqual(Decimal("10000"), create_event.price)
            self.assertEqual(order_id, create_event.order_id)
            self.assertEqual(str(self.expected_exchange_order_id), create_event.exchange_order_id)
            self.assertEqual(leverage, create_event.leverage)
            self.assertEqual(PositionAction.CLOSE.value, create_event.position)

            self.assertTrue(
                self.is_logged(
                    "INFO",
                    f"Created {OrderType.LIMIT.name} {TradeType.SELL.name} order {order_id} for "
                    f"{Decimal('100.000000')} to {PositionAction.CLOSE.name} a {self.trading_pair} position "
                    f"at {Decimal('10000.0000')}."
                )
            )

        @aioresponses()
        def test_update_order_status_when_filled(self, mock_api):
            self.exchange._set_current_timestamp(1640780000)
            request_sent_event = asyncio.Event()

            leverage = 2
            self.exchange._perpetual_trading.set_leverage(self.trading_pair, leverage)
            self.exchange.start_tracking_order(
                order_id=self.client_order_id_prefix + "1",
                exchange_order_id=self.exchange_order_id_prefix + "1",
                trading_pair=self.trading_pair,
                order_type=OrderType.LIMIT,
                trade_type=TradeType.BUY,
                price=Decimal("10000"),
                amount=Decimal("1"),
                position_action=PositionAction.OPEN,
            )
            order: InFlightOrder = self.exchange.in_flight_orders[self.client_order_id_prefix + "1"]

            urls = self.configure_completely_filled_order_status_response(
                order=order,
                mock_api=mock_api,
                callback=lambda *args, **kwargs: request_sent_event.set())

            if self.is_order_fill_http_update_included_in_status_update:
                trade_url = self.configure_full_fill_trade_response(
                    order=order,
                    mock_api=mock_api)
            else:
                # If the fill events will not be requested with the order status, we need to manually set the event
                # to allow the ClientOrderTracker to process the last status update
                order.completely_filled_event.set()
            self.async_run_with_timeout(self.exchange._update_order_status())
            # Execute one more synchronization to ensure the async task that processes the update is finished
            self.async_run_with_timeout(request_sent_event.wait())

            for url in (urls if isinstance(urls, list) else [urls]):
                order_status_request = self._all_executed_requests(mock_api, url)[0]
                self.validate_auth_credentials_present(order_status_request)
                self.validate_order_status_request(order=order, request_call=order_status_request)

            self.async_run_with_timeout(order.wait_until_completely_filled())
            self.assertTrue(order.is_done)

            if self.is_order_fill_http_update_included_in_status_update:
                self.assertTrue(order.is_filled)

                if trade_url:
                    trades_request = self._all_executed_requests(mock_api, trade_url)[0]
                    self.validate_auth_credentials_present(trades_request)
                    self.validate_trades_request(
                        order=order,
                        request_call=trades_request)

                fill_event: OrderFilledEvent = self.order_filled_logger.event_log[0]
                self.assertEqual(self.exchange.current_timestamp, fill_event.timestamp)
                self.assertEqual(order.client_order_id, fill_event.order_id)
                self.assertEqual(order.trading_pair, fill_event.trading_pair)
                self.assertEqual(order.trade_type, fill_event.trade_type)
                self.assertEqual(order.order_type, fill_event.order_type)
                self.assertEqual(order.price, fill_event.price)
                self.assertEqual(order.amount, fill_event.amount)
                self.assertEqual(self.expected_fill_fee, fill_event.trade_fee)
                self.assertEqual(leverage, fill_event.leverage)
                self.assertEqual(PositionAction.OPEN.value, fill_event.position)

            buy_event: BuyOrderCompletedEvent = self.buy_order_completed_logger.event_log[0]
            self.assertEqual(self.exchange.current_timestamp, buy_event.timestamp)
            self.assertEqual(order.client_order_id, buy_event.order_id)
            self.assertEqual(order.base_asset, buy_event.base_asset)
            self.assertEqual(order.quote_asset, buy_event.quote_asset)
            self.assertEqual(
                order.amount if self.is_order_fill_http_update_included_in_status_update else Decimal(0),
                buy_event.base_asset_amount)
            self.assertEqual(
                order.amount * order.price
                if self.is_order_fill_http_update_included_in_status_update
                else Decimal(0),
                buy_event.quote_asset_amount)
            self.assertEqual(order.order_type, buy_event.order_type)
            self.assertEqual(order.exchange_order_id, buy_event.exchange_order_id)
            self.assertNotIn(order.client_order_id, self.exchange.in_flight_orders)
            self.assertTrue(
                self.is_logged(
                    "INFO",
                    f"BUY order {order.client_order_id} completely filled."
                )
            )

        @aioresponses()
        def test_user_stream_update_for_order_full_fill(self, mock_api):
            self.exchange._set_current_timestamp(1640780000)
            leverage = 2
            self.exchange._perpetual_trading.set_leverage(self.trading_pair, leverage)
            self.exchange.start_tracking_order(
                order_id=self.client_order_id_prefix + "1",
                exchange_order_id=self.exchange_order_id_prefix + "1",
                trading_pair=self.trading_pair,
                order_type=OrderType.LIMIT,
                trade_type=TradeType.SELL,
                price=Decimal("10000"),
                amount=Decimal("1"),
                position_action=PositionAction.OPEN,
            )
            order = self.exchange.in_flight_orders[self.client_order_id_prefix + "1"]

            order_event = self.order_event_for_full_fill_websocket_update(order=order)
            trade_event = self.trade_event_for_full_fill_websocket_update(order=order)
            expected_unrealized_pnl = 12
            position_event = self.position_event_for_full_fill_websocket_update(
                order=order, unrealized_pnl=expected_unrealized_pnl
            )

            mock_queue = AsyncMock()
            event_messages = []
            if trade_event:
                event_messages.append(trade_event)
            if order_event:
                event_messages.append(order_event)
            if position_event:
                event_messages.append(position_event)
            event_messages.append(asyncio.CancelledError)
            mock_queue.get.side_effect = event_messages
            self.exchange._user_stream_tracker._user_stream = mock_queue

            if self.is_order_fill_http_update_executed_during_websocket_order_event_processing:
                self.configure_full_fill_trade_response(
                    order=order,
                    mock_api=mock_api)

            try:
                self.async_run_with_timeout(self.exchange._user_stream_event_listener())
            except asyncio.CancelledError:
                pass
            # Execute one more synchronization to ensure the async task that processes the update is finished
            self.async_run_with_timeout(order.wait_until_completely_filled())

            fill_event: OrderFilledEvent = self.order_filled_logger.event_log[0]
            self.assertEqual(self.exchange.current_timestamp, fill_event.timestamp)
            self.assertEqual(order.client_order_id, fill_event.order_id)
            self.assertEqual(order.trading_pair, fill_event.trading_pair)
            self.assertEqual(order.trade_type, fill_event.trade_type)
            self.assertEqual(order.order_type, fill_event.order_type)
            self.assertEqual(order.price, fill_event.price)
            self.assertEqual(order.amount, fill_event.amount)
            expected_fee = self.expected_fill_fee
            self.assertEqual(expected_fee, fill_event.trade_fee)
            self.assertEqual(leverage, fill_event.leverage)
            self.assertEqual(PositionAction.OPEN.value, fill_event.position)

            sell_event: SellOrderCompletedEvent = self.sell_order_completed_logger.event_log[0]
            self.assertEqual(self.exchange.current_timestamp, sell_event.timestamp)
            self.assertEqual(order.client_order_id, sell_event.order_id)
            self.assertEqual(order.base_asset, sell_event.base_asset)
            self.assertEqual(order.quote_asset, sell_event.quote_asset)
            self.assertEqual(order.amount, sell_event.base_asset_amount)
            self.assertEqual(order.amount * fill_event.price, sell_event.quote_asset_amount)
            self.assertEqual(order.order_type, sell_event.order_type)
            self.assertEqual(order.exchange_order_id, sell_event.exchange_order_id)
            self.assertNotIn(order.client_order_id, self.exchange.in_flight_orders)
            self.assertTrue(order.is_filled)
            self.assertTrue(order.is_done)

            self.assertTrue(
                self.is_logged(
                    "INFO",
                    f"SELL order {order.client_order_id} completely filled."
                )
            )

            self.assertEqual(1, len(self.exchange.account_positions))

            position: Position = self.exchange.account_positions[self.trading_pair]
            self.assertEqual(self.trading_pair, position.trading_pair)
            self.assertEqual(PositionSide.SHORT, position.position_side)
            self.assertEqual(expected_unrealized_pnl, position.unrealized_pnl)
            self.assertEqual(fill_event.price, position.entry_price)
            self.assertEqual(-fill_event.amount, position.amount)
            self.assertEqual(leverage, position.leverage)

        def test_supported_position_modes(self):
            supported_modes = self.exchange.supported_position_modes()
            self.assertEqual(self.expected_supported_position_modes, supported_modes)

        @aioresponses()
        def test_set_position_mode_failure(self, mock_api):
            request_sent_event = asyncio.Event()
            _, error_msg = self.configure_failed_set_position_mode(
                position_mode=PositionMode.HEDGE,
                mock_api=mock_api,
                callback=lambda *args, **kwargs: request_sent_event.set(),
            )
            self.exchange.set_position_mode(PositionMode.HEDGE)
            self.async_run_with_timeout(request_sent_event.wait())

            self.assertTrue(
                self.is_logged(
                    log_level="NETWORK",
                    message=f"Error switching {self.trading_pair} mode to {PositionMode.HEDGE}: {error_msg}"
                )
            )

        @aioresponses()
        def test_set_position_mode_success(self, mock_api):
            request_sent_event = asyncio.Event()
            self.configure_successful_set_position_mode(
                position_mode=PositionMode.HEDGE,
                mock_api=mock_api,
                callback=lambda *args, **kwargs: request_sent_event.set(),
            )
            self.exchange.set_position_mode(PositionMode.HEDGE)
            self.async_run_with_timeout(request_sent_event.wait())

            self.assertTrue(
                self.is_logged(
                    log_level="DEBUG",
                    message=f"Position mode switched to {PositionMode.HEDGE}.",
                )
            )

            request_sent_event.clear()
            self.configure_successful_set_position_mode(
                position_mode=PositionMode.ONEWAY,
                mock_api=mock_api,
                callback=lambda *args, **kwargs: request_sent_event.set(),
            )
            self.exchange.set_position_mode(PositionMode.ONEWAY)
            self.async_run_with_timeout(request_sent_event.wait())

            self.assertTrue(
                self.is_logged(
                    log_level="DEBUG",
                    message=f"Position mode switched to {PositionMode.ONEWAY}.",
                )
            )

        @aioresponses()
        def test_set_leverage_failure(self, mock_api):
            request_sent_event = asyncio.Event()
            target_leverage = 2
            _, message = self.configure_failed_set_leverage(
                leverage=target_leverage,
                mock_api=mock_api,
                callback=lambda *args, **kwargs: request_sent_event.set(),
            )
            self.exchange.set_leverage(trading_pair=self.trading_pair, leverage=target_leverage)
            self.async_run_with_timeout(request_sent_event.wait())

            self.assertTrue(
                self.is_logged(
                    log_level="NETWORK",
                    message=f"Error setting leverage {target_leverage} for {self.trading_pair}: {message}",
                )
            )

        @aioresponses()
        def test_set_leverage_success(self, mock_api):
            request_sent_event = asyncio.Event()
            target_leverage = 2
            self.configure_successful_set_leverage(
                leverage=target_leverage,
                mock_api=mock_api,
                callback=lambda *args, **kwargs: request_sent_event.set(),
            )
            self.exchange.set_leverage(trading_pair=self.trading_pair, leverage=target_leverage)
            self.async_run_with_timeout(request_sent_event.wait())

            self.assertTrue(
                self.is_logged(
                    log_level="INFO",
                    message=f"Leverage for {self.trading_pair} successfully set to {target_leverage}.",
                )
            )

        @aioresponses()
        @patch("asyncio.Queue.get")
        def test_listen_for_funding_info_update_initializes_funding_info(self, mock_api, mock_queue_get):
            url = self.funding_info_url

            response = self.funding_info_mock_response
            mock_api.get(url, body=json.dumps(response))

            event_messages = [asyncio.CancelledError]
            mock_queue_get.side_effect = event_messages

            try:
                self.async_run_with_timeout(self.exchange._listen_for_funding_info())
            except asyncio.CancelledError:
                pass

            funding_info: FundingInfo = self.exchange.get_funding_info(self.trading_pair)

            self.assertEqual(self.trading_pair, funding_info.trading_pair)
            self.assertEqual(self.target_funding_info_index_price, funding_info.index_price)
            self.assertEqual(self.target_funding_info_mark_price, funding_info.mark_price)
            self.assertEqual(
                self.target_funding_info_next_funding_utc_timestamp, funding_info.next_funding_utc_timestamp
            )
            self.assertEqual(self.target_funding_info_rate, funding_info.rate)

        @aioresponses()
        @patch("asyncio.Queue.get")
        def test_listen_for_funding_info_update_updates_funding_info(self, mock_api, mock_queue_get):
            url = self.funding_info_url

            response = self.funding_info_mock_response
            mock_api.get(url, body=json.dumps(response))

            funding_info_event = self.funding_info_event_for_websocket_update()

            event_messages = [funding_info_event, asyncio.CancelledError]
            mock_queue_get.side_effect = event_messages

            try:
                self.async_run_with_timeout(
                    self.exchange._listen_for_funding_info())
            except asyncio.CancelledError:
                pass

            self.assertEqual(1, self.exchange._perpetual_trading.funding_info_stream.qsize())  # rest in OB DS tests

        @aioresponses()
        def test_funding_payment_polling_loop_sends_update_event(self, mock_api):
            def callback(*args, **kwargs):
                request_sent_event.set()

            self._simulate_trading_rules_initialized()
            request_sent_event = asyncio.Event()
            url = self.funding_payment_url

            async def run_test():
                response = self.empty_funding_payment_mock_response
                mock_api.get(url, body=json.dumps(response), callback=callback)
                _ = asyncio.create_task(self.exchange._funding_payment_polling_loop())

                # Allow task to start - on first pass no event is emitted (initialization)
                await asyncio.sleep(0.1)
                self.assertEqual(0, len(self.funding_payment_logger.event_log))

                response = self.funding_payment_mock_response
                mock_api.get(url, body=json.dumps(response), callback=callback, repeat=True)

                request_sent_event.clear()
                self.exchange._funding_fee_poll_notifier.set()
                await request_sent_event.wait()
                self.assertEqual(1, len(self.funding_payment_logger.event_log))

                request_sent_event.clear()
                self.exchange._funding_fee_poll_notifier.set()
                await request_sent_event.wait()

            self.async_run_with_timeout(run_test())

            self.assertEqual(1, len(self.funding_payment_logger.event_log))
            funding_event: FundingPaymentCompletedEvent = self.funding_payment_logger.event_log[0]
            self.assertEqual(self.target_funding_payment_timestamp, funding_event.timestamp)
            self.assertEqual(self.exchange.name, funding_event.market)
            self.assertEqual(self.trading_pair, funding_event.trading_pair)
            self.assertEqual(self.target_funding_payment_payment_amount, funding_event.amount)
            self.assertEqual(self.target_funding_payment_funding_rate, funding_event.funding_rate)

        @abstractmethod
        def test_get_buy_and_sell_collateral_tokens(self):
            raise NotImplementedError



from hummingbot.strategy_v2.controllers.controller_base import ControllerBase, ControllerConfigBase
from hummingbot.strategy_v2.controllers.directional_trading_controller_base import (
    DirectionalTradingControllerBase,
    DirectionalTradingControllerConfigBase,
)
from hummingbot.strategy_v2.controllers.market_making_controller_base import (
    MarketMakingControllerBase,
    MarketMakingControllerConfigBase,
)

__all__ = [
    "ControllerBase",
    "ControllerConfigBase",
    "DirectionalTradingControllerBase",
    "DirectionalTradingControllerConfigBase",
    "MarketMakingControllerBase",
    "MarketMakingControllerConfigBase",
]



import asyncio
from abc import abstractmethod
from collections import defaultdict
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from bidict import bidict

from hummingbot.connector.constants import s_decimal_NaN
from hummingbot.connector.exchange_py_base import ExchangePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utilities.oms_connector import (
    oms_connector_constants as CONSTANTS,
    oms_connector_web_utils as ap_web_utils,
)
from hummingbot.connector.utilities.oms_connector.oms_connector_api_order_book_data_source import (
    OMSConnectorAPIOrderBookDataSource,
)
from hummingbot.connector.utilities.oms_connector.oms_connector_api_user_stream_data_source import (
    OMSConnectorAPIUserStreamDataSource,
)
from hummingbot.connector.utilities.oms_connector.oms_connector_auth import OMSConnectorAuth
from hummingbot.connector.utilities.oms_connector.oms_connector_utils import is_exchange_information_valid
from hummingbot.connector.utilities.oms_connector.oms_connector_web_utils import (
    OMSConnectorURLCreatorBase,
    OMSConnectorWebAssistantsFactory,
)
from hummingbot.connector.utils import combine_to_hb_trading_pair, get_new_numeric_client_order_id
from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderState, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.core.utils.tracking_nonce import NonceCreator
from hummingbot.core.web_assistant.connections.data_types import RESTMethod

if TYPE_CHECKING:
    from hummingbot.client.config.config_helpers import ClientConfigAdapter


class OMSExchange(ExchangePyBase):

    web_utils = ap_web_utils

    def __init__(
        self,
        client_config_map: "ClientConfigAdapter",
        api_key: str,
        secret_key: str,
        user_id: int,
        trading_pairs: Optional[List[str]] = None,
        trading_required: bool = True,
        url_creator: Optional[OMSConnectorURLCreatorBase] = None,
    ):
        self._api_key = api_key
        self._secret_key = secret_key
        self._user_id = user_id
        self._auth: Optional[OMSConnectorAuth] = None
        self._url_creator = url_creator
        self._nonce_creator = NonceCreator.for_seconds()
        self._trading_pairs = trading_pairs
        self._trading_required = trading_required
        self._web_assistants_factory: OMSConnectorWebAssistantsFactory
        self._token_id_map: Dict[int, str] = {}
        self._order_not_found_on_cancel_record: Dict[str, int] = defaultdict(lambda: 0)
        super().__init__(client_config_map)

    @property
    @abstractmethod
    def oms_id(self) -> int:
        raise NotImplementedError

    @property
    def authenticator(self) -> OMSConnectorAuth:
        if self._auth is None:
            self._auth = OMSConnectorAuth(self._api_key, self._secret_key, self._user_id)
        return self._auth

    @property
    def rate_limits_rules(self) -> List[RateLimit]:
        return CONSTANTS.RATE_LIMITS

    @property
    def client_order_id_max_length(self) -> int:
        return -1

    @property
    def client_order_id_prefix(self) -> str:
        return ""

    @property
    def trading_rules_request_path(self) -> str:
        return CONSTANTS.REST_PRODUCTS_ENDPOINT

    @property
    def trading_pairs_request_path(self) -> str:
        return CONSTANTS.REST_PRODUCTS_ENDPOINT

    @property
    def check_network_request_path(self) -> str:
        return CONSTANTS.REST_PING_ENDPOINT

    @property
    def trading_pairs(self) -> List[str]:
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return True

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    def supported_order_types(self):
        return [OrderType.LIMIT]

    async def start_network(self):
        await self._authenticate()
        await super().start_network()

    def buy(self,
            trading_pair: str,
            amount: Decimal,
            order_type=OrderType.LIMIT,
            price: Decimal = s_decimal_NaN,
            **kwargs) -> str:
        """
        Creates a promise to create a buy order using the parameters

        :param trading_pair: the token pair to operate with
        :param amount: the order amount
        :param order_type: the type of order to create (MARKET, LIMIT, LIMIT_MAKER)
        :param price: the order price

        :return: the id assigned by the connector to the order (the client id)
        """
        order_id = str(
            get_new_numeric_client_order_id(
                nonce_creator=self._nonce_creator, max_id_bit_count=CONSTANTS.MAX_ID_BIT_COUNT
            )
        )
        safe_ensure_future(self._create_order(
            trade_type=TradeType.BUY,
            order_id=order_id,
            trading_pair=trading_pair,
            amount=amount,
            order_type=order_type,
            price=price))
        return order_id

    def sell(self,
             trading_pair: str,
             amount: Decimal,
             order_type: OrderType = OrderType.LIMIT,
             price: Decimal = s_decimal_NaN,
             **kwargs) -> str:
        """
        Creates a promise to create a sell order using the parameters.
        :param trading_pair: the token pair to operate with
        :param amount: the order amount
        :param order_type: the type of order to create (MARKET, LIMIT, LIMIT_MAKER)
        :param price: the order price
        :return: the id assigned by the connector to the order (the client id)
        """
        order_id = str(
            get_new_numeric_client_order_id(
                nonce_creator=self._nonce_creator, max_id_bit_count=CONSTANTS.MAX_ID_BIT_COUNT
            )
        )
        safe_ensure_future(self._create_order(
            trade_type=TradeType.SELL,
            order_id=order_id,
            trading_pair=trading_pair,
            amount=amount,
            order_type=order_type,
            price=price))
        return order_id

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        # Not required for OMS connectors
        return False

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        # TODO: implement this method correctly for the connector
        # The default implementation was added when the functionality to detect not found orders was introduced in the
        # ExchangePyBase class. Also fix the unit test test_lost_order_removed_if_not_found_during_order_status_update
        # when replacing the dummy implementation
        return False

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        # TODO: implement this method correctly for the connector
        # The default implementation was added when the functionality to detect not found orders was introduced in the
        # ExchangePyBase class. Also fix the unit test test_cancel_order_not_found_in_the_exchange when replacing the
        # dummy implementation
        return False

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        """
        This implementation specific function is called by _cancel, and returns True if successful
        """
        start_ts = self.current_timestamp
        self.logger().debug(f"Starting cancelation of {tracked_order.client_order_id} at {start_ts}")
        await self._ensure_authenticated()
        params = {
            CONSTANTS.OMS_ID_FIELD: self.oms_id,
            CONSTANTS.ACCOUNT_ID_FIELD: self._auth.account_id,
            CONSTANTS.CL_ORDER_ID_FIELD: int(tracked_order.client_order_id),
        }
        cancel_result = await self._api_post(
            path_url=CONSTANTS.REST_ORDER_CANCELATION_ENDPOINT,
            data=params,
            is_auth_required=True,
        )
        self.logger().debug(f"Cancelation result of {tracked_order.client_order_id} at {start_ts}: {cancel_result}")
        cancel_success = False
        if cancel_result.get(CONSTANTS.ERROR_CODE_FIELD):
            if cancel_result[CONSTANTS.ERROR_CODE_FIELD] == CONSTANTS.RESOURCE_NOT_FOUND_ERR_CODE:
                await self._order_tracker.process_order_not_found(order_id)
            else:
                raise IOError(cancel_result[CONSTANTS.ERROR_MSG_FIELD])
        cancel_success = cancel_success or cancel_result[CONSTANTS.RESULT_FIELD]

        if not cancel_success:
            self.logger().debug(
                f"Failure to cancel {tracked_order.client_order_id}, attempted at {start_ts}: {cancel_result}"
            )
        elif order_id in self._order_not_found_on_cancel_record:
            del self._order_not_found_on_cancel_record[order_id]

        self.logger().debug(
            f"Cancelation of {tracked_order.client_order_id} at {start_ts} success"
        )

        return cancel_success

    def _get_fee(
        self,
        base_currency: str,
        quote_currency: str,
        order_type: OrderType,
        order_side: TradeType,
        amount: Decimal,
        price: Decimal = s_decimal_NaN,
        is_maker: Optional[bool] = None,
    ) -> TradeFeeBase:
        is_maker = False
        fee = build_trade_fee(
            self.name,
            is_maker,
            base_currency=base_currency,
            quote_currency=quote_currency,
            order_type=order_type,
            order_side=order_side,
            amount=amount,
            price=price,
        )
        return fee

    async def _place_order(
        self,
        order_id: str,
        trading_pair: str,
        amount: Decimal,
        trade_type: TradeType,
        order_type: OrderType,
        price: Decimal,
    ) -> Tuple[str, float]:
        instrument_id = await self.exchange_symbol_associated_to_pair(trading_pair)
        data = {
            CONSTANTS.INSTRUMENT_ID_FIELD: int(instrument_id),
            CONSTANTS.OMS_ID_FIELD: self.oms_id,
            CONSTANTS.ACCOUNT_ID_FIELD: self._auth.account_id,
            CONSTANTS.TIME_IN_FORCE_FIELD: CONSTANTS.GTC_TIF,
            CONSTANTS.CLIENT_ORDER_ID_FIELD: int(order_id),
            CONSTANTS.SIDE_FIELD: CONSTANTS.BUY_ACTION if trade_type == TradeType.BUY else CONSTANTS.SELL_ACTION,
            CONSTANTS.QUANTITY_FIELD: float(amount),
            CONSTANTS.ORDER_TYPE_FIELD: CONSTANTS.ORDER_TYPES[order_type],
            CONSTANTS.LIMIT_PRICE_FIELD: float(price),
        }

        send_order_resp = await self._api_request(
            path_url=CONSTANTS.REST_ORDER_CREATION_ENDPOINT,
            method=RESTMethod.POST,
            data=data,
            is_auth_required=True,
            limit_id=CONSTANTS.REST_ORDER_CREATION_ENDPOINT,
        )
        if send_order_resp.get(CONSTANTS.ERROR_CODE_FIELD):
            raise IOError(f"Error submitting order {order_id}: {send_order_resp[CONSTANTS.ERROR_MSG_FIELD]}")
        return str(send_order_resp[CONSTANTS.ORDER_ID_FIELD]), self.current_timestamp

    async def _update_trading_fees(self):
        pass

    async def _user_stream_event_listener(self):
        async for event_message in self._iter_user_event_queue():
            try:
                endpoint = event_message[CONSTANTS.MSG_ENDPOINT_FIELD]
                payload = event_message[CONSTANTS.MSG_DATA_FIELD]

                if endpoint == CONSTANTS.WS_ACC_POS_EVENT:
                    self._process_account_position_event(payload)
                elif endpoint == CONSTANTS.WS_ORDER_STATE_EVENT:
                    order = self._order_tracker.all_updatable_orders.get(str(payload[CONSTANTS.CLIENT_ORDER_ID_FIELD]))
                    if order is not None:
                        order_update = self._create_order_update(order_msg=payload, order=order)
                        self._order_tracker.process_order_update(order_update)
                elif endpoint == CONSTANTS.WS_ORDER_TRADE_EVENT:
                    order = self._order_tracker.all_fillable_orders.get(str(payload[CONSTANTS.CLIENT_ORDER_ID_FIELD]))
                    if order is not None:
                        trade_update = self._create_trade_update(trade_event=payload, order=order)
                        self._order_tracker.process_trade_update(trade_update)
                elif endpoint == CONSTANTS.WS_CANCEL_ORDER_REJECTED_EVENT:
                    pass
                else:
                    self.logger().debug(f"Unknown event received from the connector ({event_message})")
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error in user stream listener loop.")
                await self._sleep(5.0)

    async def _update_trading_rules(self):
        # This has to be reimplemented because the request requires an extra parameter
        exchange_info = await self._api_get(
            path_url=self.trading_rules_request_path,
            params={CONSTANTS.OMS_ID_FIELD: self.oms_id},
        )
        trading_rules_list = await self._format_trading_rules(exchange_info)
        self._trading_rules.clear()
        for trading_rule in trading_rules_list:
            self._trading_rules[trading_rule.trading_pair] = trading_rule
        self._initialize_trading_pair_symbols_from_exchange_info(exchange_info=exchange_info)

    async def _format_trading_rules(self, raw_trading_pair_info: List[Dict[str, Any]]):
        trading_rules = []

        for info in raw_trading_pair_info:
            try:
                if is_exchange_information_valid(exchange_info=info):
                    instrument_id = info[CONSTANTS.INSTRUMENT_ID_FIELD]
                    trading_pair = await self.trading_pair_associated_to_exchange_symbol(str(instrument_id))
                    trading_rules.append(
                        TradingRule(
                            trading_pair=trading_pair,
                            min_order_size=Decimal(str(info[CONSTANTS.MIN_QUANT_FIELD])),
                            min_price_increment=Decimal(str(info[CONSTANTS.MIN_PRICE_INCR_FIELD])),
                            min_base_amount_increment=Decimal(str(info[CONSTANTS.MIN_QUANT_INCR_FIELD])),
                        )
                    )
            except Exception:
                self.logger().exception(f"Error parsing the trading pair rule {info}. Skipping.")

        return trading_rules

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        instrument_id = await self.exchange_symbol_associated_to_pair(trading_pair)
        params = {
            CONSTANTS.OMS_ID_FIELD: self.oms_id,
            CONSTANTS.INSTRUMENT_ID_FIELD: instrument_id,
        }
        response = await self._api_request(
            path_url=CONSTANTS.REST_GET_L1_ENDPOINT, params=params
        )
        return response[CONSTANTS.LAST_TRADED_PRICE_FIELD]

    async def _update_balances(self):
        local_asset_names = set(self._account_balances.keys())
        remote_asset_names = set()

        await self._ensure_authenticated()

        params = {
            CONSTANTS.OMS_ID_FIELD: self.oms_id,
            CONSTANTS.ACCOUNT_ID_FIELD: self._auth.account_id,
        }
        account_positions: List[Dict[str, Any]] = await self._api_request(
            path_url=CONSTANTS.REST_ACC_POSITIONS_ENDPOINT,
            params=params,
            is_auth_required=True,
        )
        for position in account_positions:
            self._process_account_position_event(position)
            token = position[CONSTANTS.PRODUCT_SYMBOL_FIELD]
            remote_asset_names.add(token)

        asset_names_to_remove = local_asset_names.difference(remote_asset_names)
        for asset_name in asset_names_to_remove:
            del self._account_available_balances[asset_name]
            del self._account_balances[asset_name]

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        trade_updates = []

        if order.exchange_order_id is not None:
            exchange_order_id = int(order.exchange_order_id)
            params = {
                CONSTANTS.OMS_ID_FIELD: self.oms_id,
                CONSTANTS.ACCOUNT_ID_FIELD: self._auth.account_id,
                CONSTANTS.USER_ID_FIELD: self._auth.user_id,
                CONSTANTS.ORDER_ID_FIELD: exchange_order_id
            }

            all_fills_response = await self._api_request(
                path_url=CONSTANTS.REST_TRADE_HISTORY_ENDPOINT,
                params=params,
                is_auth_required=True,
            )

            for trade in all_fills_response:  # trades must be handled before order status updates
                trade_update = self._create_trade_update(trade_event=trade, order=order)
                trade_updates.append(trade_update)

        return trade_updates

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        exchange_order_id = await tracked_order.get_exchange_order_id()
        params = {
            CONSTANTS.OMS_ID_FIELD: self.oms_id,
            CONSTANTS.ACCOUNT_ID_FIELD: self._auth.account_id,
            CONSTANTS.ORDER_ID_FIELD: int(exchange_order_id),
        }
        updated_order_data = await self._api_request(
            path_url=CONSTANTS.REST_ORDER_STATUS_ENDPOINT,
            params=params,
            is_auth_required=True,
        )

        order_update = self._create_order_update(order_msg=updated_order_data, order=tracked_order)

        return order_update

    async def _validate_status_responses(
        self, status_responses: List[Dict[str, Any]], associated_orders: List[InFlightOrder]
    ) -> List[Dict[str, Any]]:
        validated_responses: List[Dict[str, Any]] = []
        for resp, order in zip(status_responses, associated_orders):
            if resp.get(CONSTANTS.ERROR_CODE_FIELD):
                self.logger().error(f"Error fetching order status. Response: {resp}")
                await self._order_tracker.process_order_not_found(order.client_order_id)
            else:
                validated_responses.append(resp)
        return validated_responses

    def _process_account_position_event(self, account_position_event: Dict[str, Any]):
        token = account_position_event[CONSTANTS.PRODUCT_SYMBOL_FIELD]
        amount = Decimal(str(account_position_event[CONSTANTS.AMOUNT_FIELD]))
        on_hold = Decimal(str(account_position_event[CONSTANTS.AMOUNT_ON_HOLD_FIELD]))
        self._account_balances[token] = amount
        self._account_available_balances[token] = (amount - on_hold)

    def _create_order_update(self, order_msg: Dict[str, Any], order: InFlightOrder):
        status_from_update = order_msg[CONSTANTS.ORDER_STATE_FIELD]
        if status_from_update == CONSTANTS.ACTIVE_ORDER_STATE:
            filled_amount = order_msg[CONSTANTS.QUANTITY_EXECUTED_FIELD]
            if filled_amount != 0:
                order_status = OrderState.PARTIALLY_FILLED
            else:
                order_status = OrderState.OPEN
        else:
            order_status = CONSTANTS.ORDER_STATE_MAP[status_from_update]
        order_update = OrderUpdate(
            trading_pair=order.trading_pair,
            update_timestamp=order_msg[CONSTANTS.ORDER_UPDATE_TS_FIELD] * 1e-3,
            new_state=order_status,
            client_order_id=order.client_order_id,
            exchange_order_id=str(order_msg[CONSTANTS.ORDER_ID_FIELD]),
        )
        return order_update

    def _create_trade_update(self, trade_event: Dict[str, Any], order: InFlightOrder):
        order_action = trade_event[CONSTANTS.SIDE_FIELD]
        trade_type = CONSTANTS.ORDER_SIDE_MAP[order_action]
        token_asset_id = trade_event[CONSTANTS.FEE_PRODUCT_ID_FIELD]
        fee_amount = Decimal(str(trade_event[CONSTANTS.FEE_AMOUNT_FIELD]))
        fee_token = self._token_id_map[token_asset_id] if fee_amount else order.quote_asset
        fee = TradeFeeBase.new_spot_fee(
            fee_schema=self.trade_fee_schema(),
            trade_type=trade_type,
            percent_token=fee_token,
            flat_fees=[TokenAmount(amount=fee_amount, token=fee_token)],
        )
        fill_amount = Decimal(str(trade_event[CONSTANTS.QUANTITY_FIELD.capitalize()]))
        fill_price = Decimal(str(trade_event[CONSTANTS.PRICE_FIELD]))
        trade_time = trade_event[CONSTANTS.TRADE_TIME_MS_FIELD]
        trade_update = TradeUpdate(
            trade_id=str(trade_event[CONSTANTS.TRADE_ID_FIELD]),
            client_order_id=order.client_order_id,
            exchange_order_id=str(trade_event[CONSTANTS.ORDER_ID_FIELD]),
            trading_pair=order.trading_pair,
            fee=fee,
            fill_base_amount=fill_amount,
            fill_quote_amount=fill_amount * fill_price,
            fill_price=fill_price,
            fill_timestamp=int(trade_time * 1e-3),
        )
        return trade_update

    def _create_web_assistants_factory(self) -> OMSConnectorWebAssistantsFactory:
        """We create a new authenticator to store the new session token."""
        return ap_web_utils.build_api_factory(
            throttler=self._throttler, auth=self.authenticator
        )

    async def _api_request(
        self,
        path_url,
        method: RESTMethod = RESTMethod.GET,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
        is_auth_required: bool = False,
        return_err: bool = False,
        limit_id: Optional[str] = None,
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        rest_assistant = await self._web_assistants_factory.get_rest_assistant()
        url = self._url_creator.get_rest_url(path_url)
        return await rest_assistant.execute_request(
            url=url,
            params=params,
            data=data,
            method=method,
            is_auth_required=is_auth_required,
            return_err=return_err,
            throttler_limit_id=limit_id if limit_id else path_url,
        )

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return OMSConnectorAPIOrderBookDataSource(
            self.trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            url_provider=self._url_creator,
            oms_id=self.oms_id,
        )

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return OMSConnectorAPIUserStreamDataSource(self._web_assistants_factory, self._url_creator, self.oms_id)

    async def _initialize_trading_pair_symbol_map(self):
        try:
            params = {CONSTANTS.OMS_ID_FIELD: self.oms_id}
            exchange_info = await self._api_get(path_url=self.trading_pairs_request_path, params=params)
            self._initialize_trading_pair_symbols_from_exchange_info(exchange_info=exchange_info)
        except Exception:
            self.logger().exception("There was an error requesting exchange info.")

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: List[Dict[str, Any]]):
        mapping = bidict()
        for symbol_data in filter(is_exchange_information_valid, exchange_info):
            instrument_id = str(symbol_data[CONSTANTS.INSTRUMENT_ID_FIELD])
            trading_pair = combine_to_hb_trading_pair(
                base=symbol_data[CONSTANTS.BASE_FIELD], quote=symbol_data[CONSTANTS.QUOTE_FIELD]
            )
            if instrument_id in mapping:
                self.logger().error(
                    f"Instrument ID {instrument_id} (trading pair {trading_pair}) already present in the map "
                    f"(with trading pair {mapping[instrument_id]})."
                )
                continue
            elif trading_pair in mapping.inverse:
                self.logger().error(
                    f"Trading pair {trading_pair} (instrument ID {instrument_id}) already present in the map "
                    f"(with ID {mapping.inverse[trading_pair]})."
                )
                continue
            mapping[instrument_id] = trading_pair
            base_id = symbol_data[CONSTANTS.BASE_ID_FIELD]
            base_token = symbol_data[CONSTANTS.BASE_FIELD]
            self._token_id_map[base_id] = base_token
            quote_id = symbol_data[CONSTANTS.QUOTE_ID_FIELD]
            quote_token = symbol_data[CONSTANTS.QUOTE_FIELD]
            self._token_id_map[quote_id] = quote_token
        self._set_trading_pair_symbol_map(mapping)

    async def _ensure_authenticated(self):
        if not self._auth.initialized:
            await self._authenticate()

    async def _authenticate(self):
        auth_headers = self._auth.get_rest_auth_headers()
        url = self._url_creator.get_rest_url(CONSTANTS.REST_AUTH_ENDPOINT)
        rest_assistant = await self._web_assistants_factory.get_rest_assistant()

        auth_response = await rest_assistant.execute_request(
            url,
            throttler_limit_id=CONSTANTS.REST_AUTH_ENDPOINT,
            headers=auth_headers
        )

        auth_success = self._auth.validate_rest_auth(auth_response)
        if auth_success:
            self._auth.update_with_rest_response(auth_response)
        else:
            raise IOError("Failed to authenticate.")



import sys

from hummingbot.connector.constants import SECOND
from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit
from hummingbot.core.data_type.in_flight_order import OrderState

EXCHANGE_NAME = "tegro"
DEFAULT_DOMAIN = "tegro"

DOMAIN = EXCHANGE_NAME
HBOT_ORDER_ID_PREFIX = "TEGRO-"
MAX_ORDER_ID_LEN = 32

TEGRO_BASE_URL = "https://api.tegro.com/api/"
TESTNET_BASE_URL = "https://api.testnet.tegro.com/api/"
TEGRO_WS_URL = "wss://api.tegro.com/api/v1/events/"
TESTNET_WS_URL = "wss://api.testnet.tegro.com/api/v1/events/"

DEFAULT_CHAIN = "base"
PUBLIC_WS_ENDPOINT = "ws"

# Public API endpoints or TegroClient function
TICKER_PRICE_CHANGE_PATH_URL = "v1/exchange/{}/market/{}"
EXCHANGE_INFO_PATH_LIST_URL = "v1/exchange/{}/market/list"
EXCHANGE_INFO_PATH_URL = "v1/exchange/{}/market/{}"
PING_PATH_URL = "v1/exchange/chain/list"  # TODO
SNAPSHOT_PATH_URL = "v1/orderbook/depth"
SERVER_TIME_PATH_URL = "v1/orderbook/depth"

# REST API ENDPOINTS
ACCOUNTS_PATH_URL = "v1/accounts/{}/{}/portfolio"
MARKET_LIST_PATH_URL = "v1/exchange/{}/market/list"
GENERATE_ORDER_URL = "v1/trading/market/orders/typedData/generateCancelOrder"
GENERATE_SIGN_URL = "v1/trading/market/orders/typedData/generate"
TRADES_PATH_URL = "v1/exchange/{}/market/trades"
TRADES_FOR_ORDER_PATH_URL = "v1/trading/market/orders/trades/{}"
ORDER_PATH_URL = "v1/trading/market/orders/place"
CHAIN_LIST = "v1/exchange/chain/list"
CHARTS_TRADES = "v1/exchange/{}/market/chart"
ORDER_LIST = "v1/trading/market/orders/user/{}"
CANCEL_ORDER_URL = "v1/trading/market/orders/cancel"
CANCEL_ORDER_ALL_URL = "v1/trading/market/orders/cancelAll"
TEGRO_USER_ORDER_PATH_URL = "v1/trading/market/orders/user/{}"


WS_HEARTBEAT_TIME_INTERVAL = 30

API_LIMIT_REACHED_ERROR_MESSAGE = "TOO MANY REQUESTS"
SECONDS_TO_WAIT_TO_RECEIVE_MESSAGE = 10

# Tegro params
SIDE_BUY = "buy"
SIDE_SELL = "sell"

ORDER_STATE = {
    "open": OrderState.OPEN,
    "partial": OrderState.PARTIALLY_FILLED,
    "pending": OrderState.PENDING_CANCEL,
    "completed": OrderState.FILLED,
    "cancelled": OrderState.CANCELED,
    "failed": OrderState.FAILED,
}

MAINNET_CHAIN_IDS = {
    # tegro is same as base in this case
    "base": 8453,
}

ABI = {
    "approve": [
        {
            "name": "approve",
            "stateMutability": "nonpayable",
            "type": "function",
            "inputs": [{
                "internalType": "address",
                "name": "spender",
                "type": "address"
            }, {
                "internalType": "uint256",
                "name": "value",
                "type": "uint256"
            }],
            "outputs": [{
                "internalType": "bool",
                "name": "",
                "type": "bool"
            }]
        },
    ],
    "allowance": [
        {
            "name": "allowance",
            "stateMutability": "view",
            "type": "function",
            "inputs": [{
                "internalType": "address",
                "name": "owner",
                "type": "address"
            }, {
                "internalType": "address",
                "name": "spender",
                "type": "address"
            }],
            "outputs": [{
                "internalType": "uint256",
                "name": "",
                "type": "uint256"
            }]
        }
    ]
}

Node_URLS = {
    "base": "https://mainnet.base.org",
    "tegro_base_testnet": "https://sepolia.base.org",
    "tegro_polygon_testnet": "https://rpc-amoy.polygon.technology",
    "tegro_optimism_testnet": "https://sepolia.optimism.io"
}

TESTNET_CHAIN_IDS = {
    "base": 84532,
    "polygon": 80002,
    "optimism": 11155420
}

TRADE_EVENT_TYPE = "trade_updated"
DIFF_EVENT_TYPE = "order_book_diff"

USER_METHODS = {
    "TRADES_CREATE": "user_trade_created",
    "TRADES_UPDATE": "user_trade_updated",
    "ORDER_PLACED": "order_placed",
    "ORDER_SUBMITTED": "order_submitted",
    "ORDER_TRADE_PROCESSED": "order_trade_processed"
}

HEARTBEAT_TIME_INTERVAL = 30.0

NO_LIMIT = sys.maxsize

RATE_LIMITS = [
    # Weighted Limits
    RateLimit(
        limit_id=TICKER_PRICE_CHANGE_PATH_URL,
        limit=NO_LIMIT,
        time_interval=SECOND
    ),
    RateLimit(
        limit_id=EXCHANGE_INFO_PATH_LIST_URL,
        limit=NO_LIMIT,
        time_interval=SECOND,
        linked_limits=[LinkedLimitWeightPair(TICKER_PRICE_CHANGE_PATH_URL)]
    ),
    RateLimit(
        limit_id=EXCHANGE_INFO_PATH_URL,
        limit=NO_LIMIT,
        time_interval=SECOND,
        linked_limits=[LinkedLimitWeightPair(TICKER_PRICE_CHANGE_PATH_URL)]
    ),
    RateLimit(
        limit_id=SNAPSHOT_PATH_URL,
        limit=NO_LIMIT,
        time_interval=SECOND,
        linked_limits=[LinkedLimitWeightPair(TICKER_PRICE_CHANGE_PATH_URL)]
    ),
    RateLimit(
        limit_id=SERVER_TIME_PATH_URL,
        limit=NO_LIMIT,
        time_interval=SECOND,
        linked_limits=[LinkedLimitWeightPair(TICKER_PRICE_CHANGE_PATH_URL)]
    ),
    RateLimit(
        limit_id=TEGRO_USER_ORDER_PATH_URL,
        limit=NO_LIMIT,
        time_interval=SECOND,
        linked_limits=[LinkedLimitWeightPair(TICKER_PRICE_CHANGE_PATH_URL)]
    ),
    RateLimit(
        limit_id=PING_PATH_URL,
        limit=NO_LIMIT,
        time_interval=SECOND,
        linked_limits=[LinkedLimitWeightPair(TICKER_PRICE_CHANGE_PATH_URL)]
    ),
    RateLimit(
        limit_id=ACCOUNTS_PATH_URL,
        limit=NO_LIMIT,
        time_interval=SECOND,
        linked_limits=[LinkedLimitWeightPair(TICKER_PRICE_CHANGE_PATH_URL)]
    ),
    RateLimit(
        limit_id=TRADES_PATH_URL,
        limit=NO_LIMIT,
        time_interval=SECOND,
        linked_limits=[LinkedLimitWeightPair(TICKER_PRICE_CHANGE_PATH_URL)]
    ),
    RateLimit(
        limit_id=ORDER_PATH_URL,
        limit=NO_LIMIT,
        time_interval=SECOND,
        linked_limits=[LinkedLimitWeightPair(TICKER_PRICE_CHANGE_PATH_URL)]
    ),
    RateLimit(
        limit_id=CHAIN_LIST,
        limit=NO_LIMIT,
        time_interval=SECOND,
        linked_limits=[LinkedLimitWeightPair(TICKER_PRICE_CHANGE_PATH_URL)]
    ),
    RateLimit(
        limit_id=CHARTS_TRADES,
        limit=NO_LIMIT,
        time_interval=SECOND,
        linked_limits=[LinkedLimitWeightPair(TICKER_PRICE_CHANGE_PATH_URL)]
    ),
    RateLimit(
        limit_id=ORDER_LIST,
        limit=NO_LIMIT,
        time_interval=SECOND,
        linked_limits=[LinkedLimitWeightPair(TICKER_PRICE_CHANGE_PATH_URL)]
    ),
    RateLimit(
        limit_id=MARKET_LIST_PATH_URL,
        limit=NO_LIMIT,
        time_interval=SECOND,
        linked_limits=[LinkedLimitWeightPair(TICKER_PRICE_CHANGE_PATH_URL)]
    ),
    RateLimit(
        limit_id=GENERATE_SIGN_URL,
        limit=NO_LIMIT,
        time_interval=SECOND,
        linked_limits=[LinkedLimitWeightPair(TICKER_PRICE_CHANGE_PATH_URL)]
    ),
    RateLimit(
        limit_id=GENERATE_ORDER_URL,
        limit=NO_LIMIT,
        time_interval=SECOND,
        linked_limits=[LinkedLimitWeightPair(TICKER_PRICE_CHANGE_PATH_URL)]
    ),
    RateLimit(
        limit_id=TRADES_FOR_ORDER_PATH_URL,
        limit=NO_LIMIT,
        time_interval=SECOND,
        linked_limits=[LinkedLimitWeightPair(TICKER_PRICE_CHANGE_PATH_URL)]
    ),
    RateLimit(
        limit_id=CANCEL_ORDER_URL,
        limit=NO_LIMIT,
        time_interval=SECOND,
        linked_limits=[LinkedLimitWeightPair(TICKER_PRICE_CHANGE_PATH_URL)]
    ),
    RateLimit(
        limit_id=CANCEL_ORDER_ALL_URL,
        limit=NO_LIMIT,
        time_interval=SECOND,
        linked_limits=[LinkedLimitWeightPair(TICKER_PRICE_CHANGE_PATH_URL)
                       ])
]


ORDER_NOT_EXIST_ERROR_CODE = -2013
ORDER_NOT_EXIST_MESSAGE = "Order not found"
UNKNOWN_ORDER_ERROR_CODE = -2011
UNKNOWN_ORDER_MESSAGE = "Order not found"



import hashlib
import hmac
import json
from collections import OrderedDict
from typing import Any, Dict
from urllib.parse import urlencode

from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, RESTRequest, WSRequest


class BinanceAuth(AuthBase):
    def __init__(self, api_key: str, secret_key: str, time_provider: TimeSynchronizer):
        self.api_key = api_key
        self.secret_key = secret_key
        self.time_provider = time_provider

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        """
        Adds the server time and the signature to the request, required for authenticated interactions. It also adds
        the required parameter in the request header.
        :param request: the request to be configured for authenticated interaction
        """
        if request.method == RESTMethod.POST:
            request.data = self.add_auth_to_params(params=json.loads(request.data))
        else:
            request.params = self.add_auth_to_params(params=request.params)

        headers = {}
        if request.headers is not None:
            headers.update(request.headers)
        headers.update(self.header_for_authentication())
        request.headers = headers

        return request

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        """
        This method is intended to configure a websocket request to be authenticated. Binance does not use this
        functionality
        """
        return request  # pass-through

    def add_auth_to_params(self,
                           params: Dict[str, Any]):
        timestamp = int(self.time_provider.time() * 1e3)

        request_params = OrderedDict(params or {})
        request_params["timestamp"] = timestamp

        signature = self._generate_signature(params=request_params)
        request_params["signature"] = signature

        return request_params

    def header_for_authentication(self) -> Dict[str, str]:
        return {"X-MBX-APIKEY": self.api_key}

    def _generate_signature(self, params: Dict[str, Any]) -> str:

        encoded_params_str = urlencode(params)
        digest = hmac.new(self.secret_key.encode("utf8"), encoded_params_str.encode("utf8"), hashlib.sha256).hexdigest()
        return digest



from decimal import Decimal
from typing import TYPE_CHECKING, Dict

from hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_constants import DEFAULT_DOMAIN
from hummingbot.core.rate_oracle.sources.rate_source_base import RateSourceBase
from hummingbot.core.utils import async_ttl_cache
from hummingbot.core.utils.async_utils import safe_gather

if TYPE_CHECKING:
    from hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_exchange import (
        CoinbaseAdvancedTradeExchange,
    )


class CoinbaseAdvancedTradeRateSource(RateSourceBase):
    def __init__(self):
        super().__init__()
        self._coinbase_exchange: CoinbaseAdvancedTradeExchange | None = None  # delayed because of circular reference

    @property
    def name(self) -> str:
        return "coinbase_advanced_trade"

    @async_ttl_cache(ttl=30, maxsize=1)
    async def get_prices(self, quote_token: str | None = None) -> Dict[str, Decimal]:
        if quote_token is None:
            quote_token = "USD"

        self._ensure_exchanges()
        results = {}
        tasks = [
            self._get_coinbase_prices(exchange=self._coinbase_exchange, quote_token=quote_token),
        ]
        task_results = await safe_gather(*tasks, return_exceptions=True)
        for task_result in task_results:
            if isinstance(task_result, Exception):
                self.logger().error(
                    msg="Unexpected error while retrieving rates from Coinbase. Check the log file for more info.",
                    exc_info=task_result,
                )
                break
            else:
                results |= {f"{k}-{quote_token}": v for k, v in task_result.items()}
        return results

    def _ensure_exchanges(self):
        if self._coinbase_exchange is None:
            self._coinbase_exchange = self._build_coinbase_connector_without_private_keys(domain="com")

    async def _get_coinbase_prices(
            self,
            exchange: 'CoinbaseAdvancedTradeExchange',
            quote_token: str = None) -> Dict[str, Decimal]:
        """
        Fetches coinbase prices

        :param exchange: The exchange instance from which to query prices.
        :param quote_token: A quote symbol, if specified only pairs with the quote symbol are included for prices
        :return: A dictionary of trading pairs and prices
        """
        token_price: Dict[str, str] = await exchange.get_exchange_rates(quote_token=quote_token)
        self.logger().debug(f"retrieved {len(token_price)} prices for {quote_token}")
        self.logger().debug(f"   {token_price.get('ATOM')} {quote_token} for 1 ATOM")
        return {token: Decimal(1.0) / Decimal(price) for token, price in token_price.items() if Decimal(price) != 0}

    @staticmethod
    def _build_coinbase_connector_without_private_keys(domain: str = DEFAULT_DOMAIN) -> 'CoinbaseAdvancedTradeExchange':
        from hummingbot.client.hummingbot_application import HummingbotApplication
        from hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_exchange import (
            CoinbaseAdvancedTradeExchange,
        )

        app = HummingbotApplication.main_application()
        client_config_map = app.client_config_map

        return CoinbaseAdvancedTradeExchange(
            client_config_map=client_config_map,
            coinbase_advanced_trade_api_key="",
            coinbase_advanced_trade_api_secret="",
            trading_pairs=[],
            trading_required=False,
            domain=domain,
        )



import asyncio
import logging
from typing import Any, Dict, List, Optional

from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.data_feed.coin_gecko_data_feed import coin_gecko_constants as CONSTANTS
from hummingbot.data_feed.data_feed_base import DataFeedBase
from hummingbot.logger import HummingbotLogger


class CoinGeckoDataFeed(DataFeedBase):
    cgdf_logger: Optional[HummingbotLogger] = None
    _cgdf_shared_instance: "CoinGeckoDataFeed" = None

    @classmethod
    def get_instance(cls) -> "CoinGeckoDataFeed":
        if cls._cgdf_shared_instance is None:
            cls._cgdf_shared_instance = CoinGeckoDataFeed()
        return cls._cgdf_shared_instance

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls.cgdf_logger is None:
            cls.cgdf_logger = logging.getLogger(__name__)
        return cls.cgdf_logger

    def __init__(self, update_interval: float = 30.0):
        super().__init__()
        self._ev_loop = asyncio.get_event_loop()
        self._price_dict: Dict[str, float] = {}
        self._update_interval = update_interval
        self.fetch_data_loop_task: Optional[asyncio.Task] = None
        async_throttler = AsyncThrottler(rate_limits=CONSTANTS.RATE_LIMITS)
        self._api_factory = WebAssistantsFactory(throttler=async_throttler)

    @property
    def name(self) -> str:
        return "coin_gecko_api"

    @property
    def price_dict(self) -> Dict[str, float]:
        return self._price_dict.copy()

    @property
    def health_check_endpoint(self) -> str:
        return f"{CONSTANTS.BASE_URL}{CONSTANTS.PING_REST_ENDPOINT}"

    async def start_network(self):
        await self.stop_network()
        self.fetch_data_loop_task = safe_ensure_future(self._fetch_data_loop())

    async def stop_network(self):
        if self.fetch_data_loop_task is not None:
            self.fetch_data_loop_task.cancel()
            self.fetch_data_loop_task = None

    def get_price(self, asset: str) -> float:
        return self._price_dict.get(asset.upper())

    async def get_supported_vs_tokens(self) -> List[str]:
        rest_assistant = await self._api_factory.get_rest_assistant()
        supported_vs_tokens_url = f"{CONSTANTS.BASE_URL}{CONSTANTS.SUPPORTED_VS_TOKENS_REST_ENDPOINT}"
        vs_tokens = await rest_assistant.execute_request(
            url=supported_vs_tokens_url, throttler_limit_id=CONSTANTS.REST_CALL_RATE_LIMIT_ID
        )
        return vs_tokens

    async def get_prices_by_page(
        self, vs_currency: str, page_no: int, category: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """Fetches prices specified by 250-length page. Only 50 when category is specified"""
        rest_assistant = await self._api_factory.get_rest_assistant()
        price_url: str = f"{CONSTANTS.BASE_URL}{CONSTANTS.PRICES_REST_ENDPOINT}"
        params = {
            "vs_currency": vs_currency,
            "order": "market_cap_desc",
            "per_page": 250,
            "page": page_no,
            "sparkline": "false",
        }
        if category is not None:
            params["category"] = category
        resp = await rest_assistant.execute_request(
            url=price_url, throttler_limit_id=CONSTANTS.REST_CALL_RATE_LIMIT_ID, params=params
        )
        return resp

    async def get_prices_by_token_id(self, vs_currency: str, token_ids: List[str]) -> List[Dict[str, Any]]:
        rest_assistant = await self._api_factory.get_rest_assistant()
        price_url: str = f"{CONSTANTS.BASE_URL}{CONSTANTS.PRICES_REST_ENDPOINT}"
        token_ids_str = ",".join(map(str.lower, token_ids))
        params = {
            "vs_currency": vs_currency,
            "ids": token_ids_str,
        }
        resp = await rest_assistant.execute_request(
            url=price_url, throttler_limit_id=CONSTANTS.REST_CALL_RATE_LIMIT_ID, params=params
        )
        return resp

    async def _fetch_data_loop(self):
        while True:
            try:
                await self._fetch_data()
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().network(f"Error getting data from {self.name}", exc_info=True,
                                      app_warning_msg="Couldn't fetch newest prices from Coin Gecko. "
                                                      "Check network connection.")

            await self._async_sleep(self._update_interval)

    async def _fetch_data(self):
        await self._update_asset_prices()
        self._ready_event.set()

    async def _update_asset_prices(self):
        price_dict: Dict[str, float] = {}

        for i in range(1, 5):
            try:
                results = await self.get_prices_by_page(vs_currency="usd", page_no=i)
                if 'error' in results:
                    raise Exception(f"{results['error']}")
                for result in results:
                    symbol = result["symbol"].upper()
                    price = float(result["current_price"]) if result["current_price"] is not None else 0.0
                    if symbol not in price_dict:
                        price_dict[symbol] = price
                        self._price_dict[symbol] = price
            except Exception as e:
                self.logger().warning(f"Coin Gecko API request failed. Exception: {str(e)}")
                raise e
            if i < 4:
                await self._async_sleep(0.1)
        self._price_dict = price_dict

    @staticmethod
    async def _async_sleep(delay: float):
        """Used to mock in test cases."""
        await asyncio.sleep(delay)



from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit
from hummingbot.core.data_type.in_flight_order import OrderState

DEFAULT_DOMAIN = "main"

HBOT_ORDER_ID_PREFIX = "BINGX_"
MAX_ORDER_ID_LEN = 40

SIDE_BUY = "BUY"
SIDE_SELL = "SELL"

TIME_IN_FORCE_IOC = "IOC"
TIME_IN_FORCE_POC = "POC"

REST_URLS = {"main": "https://open-api.bingx.com"}

WSS_PUBLIC_URL = {"main": "wss://open-api-ws.bingx.com/market"}

WSS_PRIVATE_URL = {"main": "wss://open-api-ws.bingx.com/market"}

# Websocket event types
DIFF_EVENT_TYPE = ""
TRADE_EVENT_TYPE = ""
SNAPSHOT_EVENT_TYPE = "depth"

# Public API endpoints
LAST_TRADED_PRICE_PATH = "/openApi/spot/v1/ticker/24hr"
EXCHANGE_INFO_PATH_URL = "/openApi/spot/v1/common/symbols"
SNAPSHOT_PATH_URL = "/openApi/spot/v1/market/depth"
SERVER_TIME_PATH_URL = "/openApi/swap/v2/server/time"

# Private API endpoints
USER_STREAM_PATH_URL = "/openApi/user/auth/userDataStream"
ACCOUNTS_PATH_URL = "/openApi/spot/v1/account/balance"
MY_TRADES_PATH_URL = "/openApi/spot/v1/trade/query"
ORDER_PATH_URL = "/openApi/spot/v1/trade/order"
CANCEL_ORDER_PATH_URL = "/openApi/spot/v1/trade/cancel"

WS_HEARTBEAT_TIME_INTERVAL = 30

# Order States
ORDER_STATE = {
    "PENDING": OrderState.PENDING_CREATE,
    "NEW": OrderState.OPEN,
    "PARTIALLY_FILLED": OrderState.PARTIALLY_FILLED,
    "FILLED": OrderState.FILLED,
    "PENDING_CANCEL": OrderState.PENDING_CANCEL,
    "CANCELED": OrderState.CANCELED,
    "FAILED": OrderState.FAILED,
}

# Rate Limit Type
REQUEST_GET = "GET"
REQUEST_GET_BURST = "GET_BURST"
REQUEST_GET_MIXED = "GET_MIXED"
REQUEST_POST = "POST"
REQUEST_POST_BURST = "POST_BURST"
REQUEST_POST_MIXED = "POST_MIXED"

# Rate Limit Max request

MAX_REQUEST_GET = 6000
MAX_REQUEST_GET_BURST = 70
MAX_REQUEST_GET_MIXED = 400
MAX_REQUEST_POST = 2400
MAX_REQUEST_POST_BURST = 50
MAX_REQUEST_POST_MIXED = 270

# Rate Limit time intervals
TWO_MINUTES = 120
ONE_SECOND = 1
SIX_SECONDS = 6
ONE_DAY = 86400

RATE_LIMITS = {
    # General
    RateLimit(limit_id=REQUEST_GET, limit=MAX_REQUEST_GET, time_interval=TWO_MINUTES),
    RateLimit(limit_id=REQUEST_GET_BURST, limit=MAX_REQUEST_GET_BURST, time_interval=ONE_SECOND),
    RateLimit(limit_id=REQUEST_GET_MIXED, limit=MAX_REQUEST_GET_MIXED, time_interval=SIX_SECONDS),
    RateLimit(limit_id=REQUEST_POST, limit=MAX_REQUEST_POST, time_interval=TWO_MINUTES),
    RateLimit(limit_id=REQUEST_POST_BURST, limit=MAX_REQUEST_POST_BURST, time_interval=ONE_SECOND),
    RateLimit(limit_id=REQUEST_POST_MIXED, limit=MAX_REQUEST_POST_MIXED, time_interval=SIX_SECONDS),
    # Linked limits
    RateLimit(limit_id=LAST_TRADED_PRICE_PATH, limit=MAX_REQUEST_GET, time_interval=TWO_MINUTES,
              linked_limits=[LinkedLimitWeightPair(REQUEST_GET, 1), LinkedLimitWeightPair(REQUEST_GET_BURST, 1),
                             LinkedLimitWeightPair(REQUEST_GET_MIXED, 1)]),
    RateLimit(limit_id=USER_STREAM_PATH_URL, limit=MAX_REQUEST_GET, time_interval=TWO_MINUTES,
              linked_limits=[LinkedLimitWeightPair(REQUEST_GET, 1), LinkedLimitWeightPair(REQUEST_GET_BURST, 1),
                             LinkedLimitWeightPair(REQUEST_GET_MIXED, 1)]),
    RateLimit(limit_id=EXCHANGE_INFO_PATH_URL, limit=MAX_REQUEST_GET, time_interval=TWO_MINUTES,
              linked_limits=[LinkedLimitWeightPair(REQUEST_GET, 1), LinkedLimitWeightPair(REQUEST_GET_BURST, 1),
                             LinkedLimitWeightPair(REQUEST_GET_MIXED, 1)]),
    RateLimit(limit_id=SNAPSHOT_PATH_URL, limit=MAX_REQUEST_GET, time_interval=TWO_MINUTES,
              linked_limits=[LinkedLimitWeightPair(REQUEST_GET, 1), LinkedLimitWeightPair(REQUEST_GET_BURST, 1),
                             LinkedLimitWeightPair(REQUEST_GET_MIXED, 1)]),
    RateLimit(limit_id=SERVER_TIME_PATH_URL, limit=MAX_REQUEST_GET, time_interval=ONE_SECOND,
              linked_limits=[LinkedLimitWeightPair(REQUEST_GET, 1), LinkedLimitWeightPair(REQUEST_GET_BURST, 1),
                             LinkedLimitWeightPair(REQUEST_GET_MIXED, 1)]),
    RateLimit(limit_id=ORDER_PATH_URL, limit=MAX_REQUEST_GET, time_interval=TWO_MINUTES,
              linked_limits=[LinkedLimitWeightPair(REQUEST_POST, 1), LinkedLimitWeightPair(REQUEST_POST_BURST, 1),
                             LinkedLimitWeightPair(REQUEST_POST_MIXED, 1)]),
    RateLimit(limit_id=CANCEL_ORDER_PATH_URL, limit=MAX_REQUEST_GET, time_interval=TWO_MINUTES,
              linked_limits=[LinkedLimitWeightPair(REQUEST_POST, 1), LinkedLimitWeightPair(REQUEST_POST_BURST, 1),
                             LinkedLimitWeightPair(REQUEST_POST_MIXED, 1)]),
    RateLimit(limit_id=ACCOUNTS_PATH_URL, limit=MAX_REQUEST_GET, time_interval=TWO_MINUTES,
              linked_limits=[LinkedLimitWeightPair(REQUEST_POST, 1), LinkedLimitWeightPair(REQUEST_POST_BURST, 1),
                             LinkedLimitWeightPair(REQUEST_POST_MIXED, 1)]),
    RateLimit(limit_id=MY_TRADES_PATH_URL, limit=MAX_REQUEST_GET, time_interval=TWO_MINUTES,
              linked_limits=[LinkedLimitWeightPair(REQUEST_POST, 1), LinkedLimitWeightPair(REQUEST_POST_BURST, 1),
                             LinkedLimitWeightPair(REQUEST_POST_MIXED, 1)]),

}


EXCHANGE_NAME = "bingx"
HBOT_BROKER_ID = "hummingbot"
HBOT_ORDER_ID = "t-HBOT"

REST_URL = "https://open-api.bingx.com/openapi"


SYMBOL_PATH_URL = "/spot/v1/common/symbols"

TRADE_EVENT_TYPE = "trade"
DIFF_EVENT_TYPE = "depth"

BINGX_USER_STREAM_PATH_URL = "/user/auth/userDataStream"
SOURCE_KEY = '7VFN4OVG'



import hashlib
import hmac
import json
from typing import Dict

from hummingbot.connector.utilities.oms_connector import oms_connector_constants as CONSTANTS
from hummingbot.core.utils.tracking_nonce import NonceCreator
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTRequest, WSJSONRequest


class OMSConnectorAuth(AuthBase):
    def __init__(self, api_key: str, secret_key: str, user_id: int):
        self.api_key = api_key
        self.secret_key = secret_key
        self.user_id = user_id
        self.user_name = None
        self.account_id = None
        self._auth_dict = None
        self._initialized = False
        self._build_auth_dict()

    @property
    def initialized(self) -> bool:
        return self._initialized

    def get_rest_auth_headers(self) -> Dict[str, str]:
        return self._auth_dict

    def validate_rest_auth(self, auth_resp_data: Dict[str, str]) -> bool:
        return self._validate_auth(auth_resp_data)

    def update_with_rest_response(self, auth_resp_data: Dict[str, str]):
        self._update_with_auth_response(auth_resp_data)

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        headers = request.headers or {}
        headers.update(self._auth_dict)
        request.headers = headers
        return request

    async def ws_authenticate(self, request: WSJSONRequest) -> WSJSONRequest:
        data = request.payload[CONSTANTS.MSG_DATA_FIELD]
        data_dict = json.loads(data)
        data_dict.update(self._auth_dict)
        request.payload[CONSTANTS.MSG_DATA_FIELD] = json.dumps(data_dict)
        return request

    def _build_auth_dict(self):
        nonce_creator = NonceCreator.for_milliseconds()
        nonce = str(nonce_creator.get_tracking_nonce())
        signature = self._generate_signature(nonce)
        self._auth_dict = {
            CONSTANTS.API_KEY_FIELD: self.api_key,
            CONSTANTS.SIGNATURE_FIELD: signature,
            CONSTANTS.USER_ID_FIELD: str(self.user_id),
            CONSTANTS.NONCE_FIELD: nonce,
        }

    def _generate_signature(self, nonce: str) -> str:
        auth_concat = f"{nonce}{self.user_id}{self.api_key}"
        signature = hmac.new(
            key=self.secret_key.encode("utf-8"),
            msg=auth_concat.encode("utf-8"),
            digestmod=hashlib.sha256,
        ).hexdigest()
        return signature

    @staticmethod
    def _validate_auth(data: Dict[str, str]) -> bool:
        return data.get(CONSTANTS.AUTHENTICATED_FIELD) or False

    def _update_with_auth_response(self, data: Dict[str, str]):
        user_data = data[CONSTANTS.USER_FIELD]
        self.user_name = user_data[CONSTANTS.USER_NAME_FIELD]
        self.account_id = user_data[CONSTANTS.ACCOUNT_ID_FIELD]
        self._initialized = True



import asyncio
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from dateutil.parser import parse as dateparse

import hummingbot.connector.exchange.btc_markets.btc_markets_constants as CONSTANTS
from hummingbot.connector.exchange.btc_markets import btc_markets_web_utils as web_utils
from hummingbot.connector.exchange.btc_markets.btc_markets_order_book import BtcMarketsOrderBook
from hummingbot.core.data_type.order_book_message import OrderBookMessage
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.btc_markets.btc_markets_exchange import BtcMarketsExchange


class BtcMarketsAPIOrderBookDataSource(OrderBookTrackerDataSource):

    _logger: Optional[HummingbotLogger] = None

    def __init__(
            self,
            trading_pairs: List[str],
            connector: 'BtcMarketsExchange',
            api_factory: WebAssistantsFactory
    ):
        super().__init__(trading_pairs)
        self._connector: BtcMarketsExchange = connector
        self._domain = CONSTANTS.DEFAULT_DOMAIN
        self._api_factory = api_factory

    async def get_last_traded_prices(self,
                                     trading_pairs: List[str],
                                     domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def _connected_websocket_assistant(self) -> WSAssistant:
        """
        Creates an instance of WSAssistant connected to the exchange

        :return: an instance of WSAssistant connected to the exchange
        """
        websocket_assistant: WSAssistant = await self._api_factory.get_ws_assistant()

        await websocket_assistant.connect(
            ws_url=CONSTANTS.WSS_V1_PUBLIC_URL[self._domain],
            ping_timeout=CONSTANTS.WS_PING_TIMEOUT)

        return websocket_assistant

    async def _subscribe_channels(self, websocket_assistant: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.

        :param websocket_assistant: the websocket assistant used to connect to the exchange
        """
        self.logger().info("Subscribing ...")
        try:
            marketIds = []
            for trading_pair in self._trading_pairs:
                symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
                marketIds.append(symbol)

            subscription_payload = {
                "messageType": "subscribe",
                "marketIds": marketIds,
                "channels": [CONSTANTS.DIFF_EVENT_TYPE, CONSTANTS.SNAPSHOT_EVENT_TYPE, CONSTANTS.TRADE_EVENT_TYPE, CONSTANTS.HEARTBEAT]
            }

            subscription_request: WSJSONRequest = WSJSONRequest(payload=subscription_payload)

            async with self._api_factory.throttler.execute_task(limit_id=CONSTANTS.WS_SUBSCRIPTION_LIMIT_ID):
                await websocket_assistant.send(subscription_request)

            self.logger().info("Subscribed to public order book and trade channels for all trading pairs ...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                "Unexpected error occurred subscribing to order book trading and delta streams...",
                exc_info=True
            )
            raise

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        """
        Identifies the channel for a particular event message. Used to find the correct queue to add the message in

        :param event_message: the event received through the websocket connection

        :return: the message channel
        """

        event_type = event_message["messageType"]
        if event_type == CONSTANTS.DIFF_EVENT_TYPE:
            return self._diff_messages_queue_key
        elif event_type == CONSTANTS.SNAPSHOT_EVENT_TYPE:
            return self._snapshot_messages_queue_key
        elif event_type == CONSTANTS.TRADE_EVENT_TYPE:
            return self._trade_messages_queue_key

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant):
        async for ws_response in websocket_assistant.iter_messages():
            data: Dict[str, Any] = ws_response.data

            channel: str = self._channel_originating_message(event_message=data)
            if channel in [self._diff_messages_queue_key, self._trade_messages_queue_key, self._snapshot_messages_queue_key]:
                self._message_queue[channel].put_nowait(data)

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        """
        Create an instance of OrderBookMessage of type OrderBookMessageType.TRADE

        :param raw_message: the JSON dictionary of the public trade event
        :param message_queue: queue where the parsed messages should be stored in
        """
        try:
            trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(raw_message["marketId"])
            timestamp: float = float(dateparse(raw_message["timestamp"]).timestamp())

            trade_message: Optional[OrderBookMessage] = BtcMarketsOrderBook.trade_message_from_exchange(
                raw_message, timestamp, {"marketId": trading_pair})

            message_queue.put_nowait(trade_message)

        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error when processing public trade updates from exchange")

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        """
        Create an instance of OrderBookMessage of type OrderBookMessageType.DIFF

        :param raw_message: the JSON dictionary of the public trade event
        :param message_queue: queue where the parsed messages should be stored in
        """
        try:
            trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(raw_message["marketId"])
            timestamp: float = float(dateparse(raw_message["timestamp"]).timestamp())

            diff_message: Optional[OrderBookMessage] = BtcMarketsOrderBook.diff_message_from_exchange(
                raw_message, timestamp, {"marketId": trading_pair})

            message_queue.put_nowait(diff_message)

        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error when processing public order book updates from exchange")

    async def _parse_order_book_snapshot_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        try:
            marketId = raw_message["marketId"]

            trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(marketId)
            timestamp: float = float(dateparse(raw_message["timestamp"]).timestamp())

            snapshot_message: Optional[OrderBookMessage] = BtcMarketsOrderBook.snapshot_message_from_exchange_rest(
                raw_message, timestamp, {"marketId": trading_pair})

            message_queue.put_nowait(snapshot_message)

        except asyncio.CancelledError:
            raise
        except Exception:
            marketId = raw_message["marketId"]
            self.logger().error(f"Unexpected error fetching order book snapshot for {marketId}.", exc_info=True)
            await self._sleep(5.0)

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        try:
            snapshot: Dict[str, Any] = await self.get_snapshot(trading_pair=trading_pair)
            snapshot_timestamp: float = float(snapshot["snapshotId"])

            return BtcMarketsOrderBook.snapshot_message_from_exchange_rest(
                snapshot,
                snapshot_timestamp,
                metadata={"marketId": trading_pair}
            )
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(f"Unexpected error fetching order book snapshot for {trading_pair}.", exc_info=True)
            await self._sleep(5.0)

    async def get_snapshot(
            self,
            trading_pair: str,
            limit: int = 1000,
    ) -> Dict[str, Any]:
        """
        Retrieves a copy of the full order book from the exchange, for a particular trading pair.
        :param trading_pair: the trading pair for which the order book will be retrieved
        :param limit: the depth of the order book to retrieve
        :return: the response from the exchange (JSON dictionary)
        """
        params = {}
        if limit != 0:
            params["limit"] = str(limit)

        ex_trading_pair = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

        rest_assistant = await self._api_factory.get_rest_assistant()
        data = await rest_assistant.execute_request(
            url=web_utils.public_rest_url(path_url=f"{CONSTANTS.MARKETS_URL}/{ex_trading_pair}/orderbook"),
            method=RESTMethod.GET,
            throttler_limit_id=CONSTANTS.MARKETS_URL,
        )

        return data



import asyncio
from typing import Any, Dict, List, Mapping, Optional

from google.protobuf import any_pb2
from pyinjective import Transaction
from pyinjective.async_client import AsyncClient
from pyinjective.composer import Composer, injective_exchange_tx_pb
from pyinjective.core.network import Network

from hummingbot.connector.exchange.injective_v2 import injective_constants as CONSTANTS
from hummingbot.connector.exchange.injective_v2.data_sources.injective_data_source import InjectiveDataSource
from hummingbot.connector.exchange.injective_v2.injective_market import (
    InjectiveDerivativeMarket,
    InjectiveSpotMarket,
    InjectiveToken,
)
from hummingbot.connector.exchange.injective_v2.injective_query_executor import PythonSDKInjectiveQueryExecutor
from hummingbot.connector.gateway.gateway_in_flight_order import GatewayInFlightOrder, GatewayPerpetualInFlightOrder
from hummingbot.connector.utils import combine_to_hb_trading_pair
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.api_throttler.async_throttler_base import AsyncThrottlerBase
from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.common import OrderType
from hummingbot.core.data_type.in_flight_order import OrderUpdate
from hummingbot.core.pubsub import PubSub
from hummingbot.logger import HummingbotLogger


class InjectiveReadOnlyDataSource(InjectiveDataSource):
    _logger: Optional[HummingbotLogger] = None

    def __init__(
            self,
            network: Network,
            rate_limits: List[RateLimit],
            use_secure_connection: bool = True):
        self._network = network
        self._client = AsyncClient(
            network=self._network,
            insecure=not use_secure_connection,
        )
        self._composer = None
        self._query_executor = PythonSDKInjectiveQueryExecutor(sdk_client=self._client)

        self._publisher = PubSub()
        self._last_received_message_timestamp = 0
        self._throttler = AsyncThrottler(rate_limits=rate_limits)

        self._markets_initialization_lock = asyncio.Lock()
        self._spot_market_info_map: Optional[Dict[str, InjectiveSpotMarket]] = None
        self._derivative_market_info_map: Optional[Dict[str, InjectiveDerivativeMarket]] = None
        self._spot_market_and_trading_pair_map: Optional[Mapping[str, str]] = None
        self._derivative_market_and_trading_pair_map: Optional[Mapping[str, str]] = None
        self._tokens_map: Optional[Dict[str, InjectiveToken]] = None
        self._token_symbol_and_denom_map: Optional[Mapping[str, str]] = None

        self._events_listening_tasks: List[asyncio.Task] = []

    @property
    def publisher(self):
        return self._publisher

    @property
    def query_executor(self):
        return self._query_executor

    @property
    def throttler(self):
        return self._throttler

    @property
    def portfolio_account_injective_address(self) -> str:
        raise NotImplementedError

    @property
    def portfolio_account_subaccount_id(self) -> str:
        return ""

    @property
    def trading_account_injective_address(self) -> str:
        raise NotImplementedError

    @property
    def injective_chain_id(self) -> str:
        return self._network.chain_id

    @property
    def fee_denom(self) -> str:
        return self._network.fee_denom

    @property
    def portfolio_account_subaccount_index(self) -> int:
        raise NotImplementedError

    @property
    def network_name(self) -> str:
        return self._network.string()

    @property
    def last_received_message_timestamp(self) -> float:
        return self._last_received_message_timestamp

    async def composer(self) -> Composer:
        if self._composer is None:
            self._composer = await self._client.composer()
        return self._composer

    async def timeout_height(self) -> int:
        raise NotImplementedError

    async def spot_market_and_trading_pair_map(self):
        if self._spot_market_and_trading_pair_map is None:
            async with self._markets_initialization_lock:
                if self._spot_market_and_trading_pair_map is None:
                    await self.update_markets()
        return self._spot_market_and_trading_pair_map.copy()

    async def spot_market_info_for_id(self, market_id: str):
        if self._spot_market_info_map is None:
            async with self._markets_initialization_lock:
                if self._spot_market_info_map is None:
                    await self.update_markets()

        return self._spot_market_info_map[market_id]

    async def derivative_market_and_trading_pair_map(self):
        if self._derivative_market_and_trading_pair_map is None:
            async with self._markets_initialization_lock:
                if self._derivative_market_and_trading_pair_map is None:
                    await self.update_markets()
        return self._derivative_market_and_trading_pair_map.copy()

    async def derivative_market_info_for_id(self, market_id: str):
        if self._derivative_market_info_map is None:
            async with self._markets_initialization_lock:
                if self._derivative_market_info_map is None:
                    await self.update_markets()

        return self._derivative_market_info_map[market_id]

    async def trading_pair_for_market(self, market_id: str):
        if self._spot_market_and_trading_pair_map is None or self._derivative_market_and_trading_pair_map is None:
            async with self._markets_initialization_lock:
                if self._spot_market_and_trading_pair_map is None or self._derivative_market_and_trading_pair_map is None:
                    await self.update_markets()

        trading_pair = self._spot_market_and_trading_pair_map.get(market_id)

        if trading_pair is None:
            trading_pair = self._derivative_market_and_trading_pair_map[market_id]
        return trading_pair

    async def market_id_for_spot_trading_pair(self, trading_pair: str) -> str:
        if self._spot_market_and_trading_pair_map is None:
            async with self._markets_initialization_lock:
                if self._spot_market_and_trading_pair_map is None:
                    await self.update_markets()

        return self._spot_market_and_trading_pair_map.inverse[trading_pair]

    async def market_id_for_derivative_trading_pair(self, trading_pair: str) -> str:
        if self._derivative_market_and_trading_pair_map is None:
            async with self._markets_initialization_lock:
                if self._derivative_market_and_trading_pair_map is None:
                    await self.update_markets()

        return self._derivative_market_and_trading_pair_map.inverse[trading_pair]

    async def spot_markets(self):
        if self._spot_market_and_trading_pair_map is None:
            async with self._markets_initialization_lock:
                if self._spot_market_and_trading_pair_map is None:
                    await self.update_markets()

        return list(self._spot_market_info_map.values())

    async def derivative_markets(self):
        if self._derivative_market_and_trading_pair_map is None:
            async with self._markets_initialization_lock:
                if self._derivative_market_and_trading_pair_map is None:
                    await self.update_markets()

        return list(self._derivative_market_info_map.values())

    async def token(self, denom: str) -> InjectiveToken:
        if self._tokens_map is None:
            async with self._markets_initialization_lock:
                if self._tokens_map is None:
                    await self.update_markets()

        return self._tokens_map.get(denom)

    def events_listening_tasks(self) -> List[asyncio.Task]:
        return self._events_listening_tasks.copy()

    def add_listening_task(self, task: asyncio.Task):
        self._events_listening_tasks.append(task)

    def configure_throttler(self, throttler: AsyncThrottlerBase):
        self._throttler = throttler

    async def trading_account_sequence(self) -> int:
        raise NotImplementedError

    async def trading_account_number(self) -> int:
        raise NotImplementedError

    async def initialize_trading_account(self):  # pragma: no cover
        # Do nothing
        pass

    async def update_markets(self):
        (
            self._tokens_map,
            self._token_symbol_and_denom_map,
            self._spot_market_info_map,
            self._spot_market_and_trading_pair_map,
            self._derivative_market_info_map,
            self._derivative_market_and_trading_pair_map,
        ) = await self._get_markets_and_tokens()

    def real_tokens_spot_trading_pair(self, unique_trading_pair: str) -> str:
        resulting_trading_pair = unique_trading_pair
        if (self._spot_market_and_trading_pair_map is not None
                and self._spot_market_info_map is not None):
            market_id = self._spot_market_and_trading_pair_map.inverse.get(unique_trading_pair)
            market = self._spot_market_info_map.get(market_id)
            if market is not None:
                resulting_trading_pair = combine_to_hb_trading_pair(
                    base=market.base_token.symbol,
                    quote=market.quote_token.symbol,
                )

        return resulting_trading_pair

    def real_tokens_perpetual_trading_pair(self, unique_trading_pair: str) -> str:
        resulting_trading_pair = unique_trading_pair
        if (self._derivative_market_and_trading_pair_map is not None
                and self._derivative_market_info_map is not None):
            market_id = self._derivative_market_and_trading_pair_map.inverse.get(unique_trading_pair)
            market = self._derivative_market_info_map.get(market_id)
            if market is not None:
                resulting_trading_pair = combine_to_hb_trading_pair(
                    base=market.base_token_symbol(),
                    quote=market.quote_token.symbol,
                )

        return resulting_trading_pair

    async def order_updates_for_transaction(
            self,
            transaction_hash: str,
            spot_orders: Optional[List[GatewayInFlightOrder]] = None,
            perpetual_orders: Optional[List[GatewayPerpetualInFlightOrder]] = None
    ) -> List[OrderUpdate]:
        raise NotImplementedError

    def supported_order_types(self) -> List[OrderType]:
        return []

    async def _initialize_timeout_height(self):  # pragma: no cover
        # Do nothing
        pass

    def _sign_and_encode(self, transaction: Transaction) -> bytes:
        raise NotImplementedError

    def _uses_default_portfolio_subaccount(self) -> bool:
        return True

    async def _order_creation_messages(
            self,
            spot_orders_to_create: List[GatewayInFlightOrder],
            derivative_orders_to_create: List[GatewayPerpetualInFlightOrder]
    ) -> List[any_pb2.Any]:
        raise NotImplementedError

    async def _order_cancel_message(
            self,
            spot_orders_to_cancel: List[injective_exchange_tx_pb.OrderData],
            derivative_orders_to_cancel: List[injective_exchange_tx_pb.OrderData]
    ) -> any_pb2.Any:
        raise NotImplementedError

    async def _all_subaccount_orders_cancel_message(
            self,
            spot_orders_to_cancel: List[injective_exchange_tx_pb.OrderData],
            derivative_orders_to_cancel: List[injective_exchange_tx_pb.OrderData]
    ) -> any_pb2.Any:
        raise NotImplementedError

    async def _generate_injective_order_data(
            self,
            order: GatewayInFlightOrder,
            market_id: str,
    ) -> injective_exchange_tx_pb.OrderData:
        raise NotImplementedError

    async def _updated_derivative_market_info_for_id(self, market_id: str) -> Dict[str, Any]:
        async with self.throttler.execute_task(limit_id=CONSTANTS.DERIVATIVE_MARKETS_LIMIT_ID):
            market_info = await self._query_executor.derivative_market(market_id=market_id)

        return market_info

    async def _process_chain_stream_update(
            self, chain_stream_update: Dict[str, Any], derivative_markets: List[InjectiveDerivativeMarket],
    ):
        self._last_received_message_timestamp = self._time()
        await super()._process_chain_stream_update(
            chain_stream_update=chain_stream_update,
            derivative_markets=derivative_markets,
        )

    async def _process_transaction_update(self, transaction_event: Dict[str, Any]):
        self._last_received_message_timestamp = self._time()
        await super()._process_transaction_update(transaction_event=transaction_event)

    async def _configure_gas_fee_for_transaction(self, transaction: Transaction):
        raise NotImplementedError



import asyncio
import logging
from decimal import Decimal
from typing import Dict, Optional, Set

from pydantic import BaseModel

from hummingbot.connector.utils import split_hb_trading_pair
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.gateway.gateway_http_client import GatewayHttpClient
from hummingbot.core.network_base import NetworkBase
from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.logger import HummingbotLogger


class TokenBuySellPrice(BaseModel):
    base: str
    quote: str
    connector: str
    chain: str
    network: str
    order_amount_in_base: Decimal
    buy_price: Decimal
    sell_price: Decimal


class AmmGatewayDataFeed(NetworkBase):
    dex_logger: Optional[HummingbotLogger] = None
    gateway_client = GatewayHttpClient.get_instance()

    def __init__(
        self,
        connector_chain_network: str,
        trading_pairs: Set[str],
        order_amount_in_base: Decimal,
        update_interval: float = 1.0,
    ) -> None:
        super().__init__()
        self._ev_loop = asyncio.get_event_loop()
        self._price_dict: Dict[str, TokenBuySellPrice] = {}
        self._update_interval = update_interval
        self.fetch_data_loop_task: Optional[asyncio.Task] = None
        # param required for DEX API request
        self.connector_chain_network = connector_chain_network
        self.trading_pairs = trading_pairs
        self.order_amount_in_base = order_amount_in_base

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls.dex_logger is None:
            cls.dex_logger = logging.getLogger(__name__)
        return cls.dex_logger

    @property
    def name(self) -> str:
        return f"AmmDataFeed[{self.connector_chain_network}]"

    @property
    def connector(self) -> str:
        return self.connector_chain_network.split("_")[0]

    @property
    def chain(self) -> str:
        return self.connector_chain_network.split("_")[1]

    @property
    def network(self) -> str:
        return self.connector_chain_network.split("_")[2]

    @property
    def price_dict(self) -> Dict[str, TokenBuySellPrice]:
        return self._price_dict

    def is_ready(self) -> bool:
        return len(self._price_dict) == len(self.trading_pairs)

    async def check_network(self) -> NetworkStatus:
        is_gateway_online = await self.gateway_client.ping_gateway()
        if not is_gateway_online:
            self.logger().warning("Gateway is not online. Please check your gateway connection.")
        return NetworkStatus.CONNECTED if is_gateway_online else NetworkStatus.NOT_CONNECTED

    async def start_network(self) -> None:
        await self.stop_network()
        self.fetch_data_loop_task = safe_ensure_future(self._fetch_data_loop())

    async def stop_network(self) -> None:
        if self.fetch_data_loop_task is not None:
            self.fetch_data_loop_task.cancel()
            self.fetch_data_loop_task = None

    async def _fetch_data_loop(self) -> None:
        while True:
            try:
                await self._fetch_data()
            except asyncio.CancelledError:
                raise
            except Exception as e:
                self.logger().error(
                    f"Error getting data from {self.name}"
                    f"Check network connection. Error: {e}",
                )
            await self._async_sleep(self._update_interval)

    async def _fetch_data(self) -> None:
        token_price_tasks = [
            asyncio.create_task(self._register_token_buy_sell_price(trading_pair))
            for trading_pair in self.trading_pairs
        ]
        await asyncio.gather(*token_price_tasks)

    async def _register_token_buy_sell_price(self, trading_pair: str) -> None:
        base, quote = split_hb_trading_pair(trading_pair)
        token_buy_price_task = asyncio.create_task(self._request_token_price(trading_pair, TradeType.BUY))
        token_sell_price_task = asyncio.create_task(self._request_token_price(trading_pair, TradeType.SELL))
        self._price_dict[trading_pair] = TokenBuySellPrice(
            base=base,
            quote=quote,
            connector=self.connector,
            chain=self.chain,
            network=self.network,
            order_amount_in_base=self.order_amount_in_base,
            buy_price=await token_buy_price_task,
            sell_price=await token_sell_price_task,
        )

    async def _request_token_price(self, trading_pair: str, trade_type: TradeType) -> Decimal:
        base, quote = split_hb_trading_pair(trading_pair)
        connector, chain, network = self.connector_chain_network.split("_")
        token_price = await self.gateway_client.get_price(
            chain,
            network,
            connector,
            base,
            quote,
            self.order_amount_in_base,
            trade_type,
        )
        return Decimal(token_price["price"])

    @staticmethod
    async def _async_sleep(delay: float) -> None:
        """Used to mock in test cases."""
        await asyncio.sleep(delay)



import sys

from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.common import OrderType
from hummingbot.core.data_type.in_flight_order import OrderState

CLIENT_ID_PREFIX = "93027a12dac34fBC"
MAX_ID_LEN = 32
SECONDS_TO_WAIT_TO_RECEIVE_MESSAGE = 30 * 0.8

DEFAULT_DOMAIN = ""

# URLs

OKX_BASE_URL = "https://www.okx.com/"

# Doesn't include base URL as the tail is required to generate the signature

OKX_SERVER_TIME_PATH = '/api/v5/public/time'
OKX_INSTRUMENTS_PATH = '/api/v5/public/instruments'
OKX_TICKER_PATH = '/api/v5/market/ticker'
OKX_TICKERS_PATH = '/api/v5/market/tickers'
OKX_ORDER_BOOK_PATH = '/api/v5/market/books'

# Auth required
OKX_PLACE_ORDER_PATH = "/api/v5/trade/order"
OKX_ORDER_DETAILS_PATH = '/api/v5/trade/order'
OKX_ORDER_CANCEL_PATH = '/api/v5/trade/cancel-order'
OKX_BATCH_ORDER_CANCEL_PATH = '/api/v5/trade/cancel-batch-orders'
OKX_BALANCE_PATH = '/api/v5/account/balance'
OKX_TRADE_FILLS_PATH = "/api/v5/trade/fills"

# WS
OKX_WS_URI_PUBLIC = "wss://ws.okx.com:8443/ws/v5/public"
OKX_WS_URI_PRIVATE = "wss://ws.okx.com:8443/ws/v5/private"

OKX_WS_ACCOUNT_CHANNEL = "account"
OKX_WS_ORDERS_CHANNEL = "orders"
OKX_WS_PUBLIC_TRADES_CHANNEL = "trades"
OKX_WS_PUBLIC_BOOKS_CHANNEL = "books"

OKX_WS_CHANNELS = {
    OKX_WS_ACCOUNT_CHANNEL,
    OKX_WS_ORDERS_CHANNEL
}

WS_CONNECTION_LIMIT_ID = "WSConnection"
WS_REQUEST_LIMIT_ID = "WSRequest"
WS_SUBSCRIPTION_LIMIT_ID = "WSSubscription"
WS_LOGIN_LIMIT_ID = "WSLogin"

ORDER_STATE = {
    "live": OrderState.OPEN,
    "filled": OrderState.FILLED,
    "partially_filled": OrderState.PARTIALLY_FILLED,
    "canceled": OrderState.CANCELED,
}

ORDER_TYPE_MAP = {
    OrderType.LIMIT: "limit",
    OrderType.MARKET: "market",
    OrderType.LIMIT_MAKER: "post_only",
}

NO_LIMIT = sys.maxsize

RATE_LIMITS = [
    RateLimit(WS_CONNECTION_LIMIT_ID, limit=3, time_interval=1),
    RateLimit(WS_REQUEST_LIMIT_ID, limit=100, time_interval=10),
    RateLimit(WS_SUBSCRIPTION_LIMIT_ID, limit=240, time_interval=60 * 60),
    RateLimit(WS_LOGIN_LIMIT_ID, limit=1, time_interval=15),
    RateLimit(limit_id=OKX_SERVER_TIME_PATH, limit=10, time_interval=2),
    RateLimit(limit_id=OKX_INSTRUMENTS_PATH, limit=20, time_interval=2),
    RateLimit(limit_id=OKX_TICKER_PATH, limit=20, time_interval=2),
    RateLimit(limit_id=OKX_TICKERS_PATH, limit=20, time_interval=2),
    RateLimit(limit_id=OKX_ORDER_BOOK_PATH, limit=20, time_interval=2),
    RateLimit(limit_id=OKX_PLACE_ORDER_PATH, limit=20, time_interval=2),
    RateLimit(limit_id=OKX_ORDER_DETAILS_PATH, limit=20, time_interval=2),
    RateLimit(limit_id=OKX_ORDER_CANCEL_PATH, limit=20, time_interval=2),
    RateLimit(limit_id=OKX_BATCH_ORDER_CANCEL_PATH, limit=300, time_interval=2),
    RateLimit(limit_id=OKX_BALANCE_PATH, limit=10, time_interval=2),
    RateLimit(limit_id=OKX_TRADE_FILLS_PATH, limit=60, time_interval=2),
]



from decimal import Decimal
from enum import Enum
from typing import List, Optional

from hummingbot.core.data_type.common import TradeType
from hummingbot.strategy_v2.executors.data_types import ExecutorConfigBase
from hummingbot.strategy_v2.executors.position_executor.data_types import TrailingStop


class DCAMode(Enum):
    MAKER = "MAKER"
    TAKER = "TAKER"


class DCAExecutorConfig(ExecutorConfigBase):
    type: str = "dca_executor"
    connector_name: str
    trading_pair: str
    side: TradeType
    leverage: int = 1
    amounts_quote: List[Decimal]
    prices: List[Decimal]
    take_profit: Optional[Decimal] = None
    stop_loss: Optional[Decimal] = None
    trailing_stop: Optional[TrailingStop] = None
    time_limit: Optional[int] = None
    mode: DCAMode = DCAMode.MAKER
    activation_bounds: Optional[List[Decimal]] = None
    level_id: Optional[str] = None



#!/usr/bin/env python

from collections import namedtuple
from decimal import Decimal
from enum import Enum


class AssetType(Enum):
    BASE_CURRENCY = 1
    QUOTE_CURRENCY = 2


class MarketConfig(namedtuple("_MarketConfig", "buy_fees_asset,"
                                               "buy_fees_amount,"
                                               "sell_fees_asset,"
                                               "sell_fees_amount,")):
    buy_fees_asset: AssetType
    buy_fees_amount: Decimal
    sell_fees_asset: AssetType
    sell_fees_amount: Decimal

    @classmethod
    def default_config(cls) -> "MarketConfig":
        return MarketConfig(AssetType.BASE_CURRENCY, Decimal(0.0), AssetType.QUOTE_CURRENCY, Decimal(0.0))

    @classmethod
    def create_config(cls, trading_fee: float):
        return MarketConfig(AssetType.BASE_CURRENCY, trading_fee, AssetType.QUOTE_CURRENCY, trading_fee)
